import pytest
from ase import Atoms
from gpaw import GPAW, Mixer, Davidson
from gpaw.test import equal

# ??? g = Generator('H', 'TPSS', scalarrel=True, nofiles=True)


@pytest.mark.mgga
def test_mgga_nsc_MGGA(in_tmp_dir):
    def xc(name):
        return dict(name=name, stencil=1)

    atoms = Atoms('H', magmoms=[1], pbc=True)
    atoms.center(vacuum=3)

    def getkwargs():
        return dict(mode='fd',
                    eigensolver=Davidson(3),
                    mixer=Mixer(0.7, 5, 50.0),
                    parallel=dict(augment_grids=True),
                    gpts=(32, 32, 32), nbands=1, xc=xc('oldPBE'))

    calc = GPAW(txt='Hnsc.txt', **getkwargs())
    atoms.calc = calc
    e1 = atoms.get_potential_energy()
    _ = calc.get_reference_energy()
    de12t = calc.get_xc_difference(xc('TPSS'))
    de12m = calc.get_xc_difference(xc('M06-L'))
    de12m2 = calc.get_xc_difference(xc('M06-L'))
    de12r = calc.get_xc_difference(xc('revTPSS'))

    print('================')
    print('e1 = ', e1)
    print('de12t = ', de12t)
    print('de12m = ', de12m)
    print('de12r = ', de12r)
    print('tpss = ', e1 + de12t)
    print('m06l = ', e1 + de12m)
    print('m06l = ', e1 + de12m2)
    print('revtpss = ', e1 + de12r)
    print('================')

    equal(e1 + de12t, -1.11723235592, 0.005)
    equal(e1 + de12m, -1.18207312133, 0.005)
    equal(e1 + de12r, -1.10093196353, 0.005)

    # ??? g = Generator('He', 'TPSS', scalarrel=True, nofiles=True)

    atomsHe = Atoms('He', pbc=True)
    atomsHe.center(vacuum=3)
    calc = GPAW(txt='Hensc.txt', **getkwargs())
    atomsHe.calc = calc
    e1He = atomsHe.get_potential_energy()
    _ = calc.get_reference_energy()
    de12tHe = calc.get_xc_difference(xc('TPSS'))
    de12mHe = calc.get_xc_difference(xc('M06-L'))
    de12rHe = calc.get_xc_difference(xc('revTPSS'))

    print('================')
    print('e1He = ', e1He)
    print('de12tHe = ', de12tHe)
    print('de12mHe = ', de12mHe)
    print('de12rHe = ', de12rHe)
    print('tpss = ', e1He + de12tHe)
    print('m06l = ', e1He + de12mHe)
    print('revtpss = ', e1He + de12rHe)
    print('================')

    equal(e1He + de12tHe, -0.409972893501, 0.005)
    equal(e1He + de12mHe, -0.487249688866, 0.005)
    equal(e1He + de12rHe, -0.447232286813, 0.005)

    energy_tolerance = 0.001
    equal(e1, -1.124, energy_tolerance)
    equal(e1He, 0.0100192, energy_tolerance)



"""Functionality to calculate the all-electron Fourier components of local
functions of the electon (spin-)density."""

from abc import ABC, abstractmethod

import numpy as np
from scipy.special import spherical_jn

from ase.units import Bohr

from gpaw.response import ResponseGroundStateAdapter, ResponseContext, timer

from gpaw.spherical_harmonics import Yarr
from gpaw.sphere.rshe import calculate_reduced_rshe
from gpaw.xc import XC
from gpaw.xc.libxc import LibXC


class LocalFTCalculator(ABC):
    r"""Calculator base class for calculators of all-electron plane-wave
    components to arbitrary real-valued real-space functionals f[n](r) which
    can be written as closed form functions of the local ground state
    (spin-)density:

    f[n](r) = f(n(r)).

    Since n(r) is lattice periodic, so is f(r) and the plane-wave components
    can be calculated as (see [PRB 103, 245110 (2021)] for definitions)

           /
    f(G) = |dr f(r) e^(-iG.r),
           /
            V0

    where V0 is the unit-cell volume.
    """

    def __init__(self, gs, context, bg_density=None):
        """Constructor for the LocalFTCalculator

        Parameters
        ----------
        gs : ResponseGroundStateAdapter
            Adapter containing relevant information about the underlying DFT
            ground state
        context : ResponseContext
        bg_density : float
            Spin-neutral background electron density (in Å^-3) to add to the
            actual electron density in order to regularize functions which
            diverge in vacuum.
        """
        assert isinstance(gs, ResponseGroundStateAdapter)
        self.gs = gs
        assert isinstance(context, ResponseContext)
        self.context = context

        if bg_density is None:
            self.bg_density = None
        else:
            assert isinstance(bg_density, float)
            # Convert to atomic units
            self.bg_density = bg_density * Bohr**3.  # Å^-3 -> Bohr^-3

    @staticmethod
    def from_rshe_parameters(gs, context, bg_density=None,
                             rshelmax=-1, rshewmin=None):
        """Construct the LocalFTCalculator based on parameters for the
        expansion of the PAW correction in real spherical harmonics

        Parameters
        ----------
        rshelmax : int or None
            Expand f(r) in real spherical harmonics inside the augmentation
            spheres. If None, the plane-wave components will be calculated
            without augmentation. The value of rshelmax indicates the maximum
            index l to perform the expansion in (l < 6).
        rshewmin : float or None
            If None, the PAW correction will be fully expanded up to the chosen
            lmax. Given as a float (0 < rshewmin < 1), rshewmin indicates what
            coefficients to use in the expansion. If any (l,m) coefficient
            contributes with less than a fraction of rshewmin on average, it
            will not be included.
        """
        if rshelmax is None:
            return LocalGridFTCalculator(gs, context, bg_density=bg_density)
        else:
            return LocalPAWFTCalculator(gs, context, bg_density=bg_density,
                                        rshelmax=rshelmax, rshewmin=rshewmin)

    @timer('LocalFT')
    def __call__(self, qpd, add_f):
        """Calculate the plane-wave components f(G).

        Parameters
        ----------
        qpd : SingleQPWDescriptor
            Defines the plane-wave basis to calculate the components in.
        add_f : method
            Defines the local function of the electron (spin-)density to
            Fourier transform. Should take arguments gd (GridDescriptor),
            n_sR (electron spin-density on the real space grid of gd) and
            f_R (output array) and add the function f(R) to the output array.
            Example:
            >>> def add_total_density(gd, n_sR, f_R):
            ...     f_R += np.sum(n_sR, axis=0)

        Returns
        -------
        f_G : np.array
            Plane-wave components of the function f, indexes by the reciprocal
            lattice vectors G.
        """
        self.context.print('Calculating f(G)')
        f_G = self.calculate(qpd, add_f)
        self.context.print('Finished calculating f(G)')

        return f_G

    @abstractmethod
    def calculate(self, qpd, add_f):
        pass

    @staticmethod
    def equivalent_real_space_grids(gd1, gd2):
        assert gd1.comm.size == 1
        assert gd2.comm.size == 1
        return (gd1.N_c == gd2.N_c).all()

    def get_electron_density(self, gd, pseudo=False):
        """Get the electron density corresponding to a given grid descriptor.
        """
        gridrefinement = self.get_gridrefinement(gd)

        if pseudo:
            _get_electron_density = self.gs.get_pseudo_density
        else:
            _get_electron_density = self.gs.get_all_electron_density

        n_sR, gdref = _get_electron_density(gridrefinement=gridrefinement)

        assert self.equivalent_real_space_grids(gd, gdref)

        if self.bg_density is not None:
            # Add spin-neutral background electron density
            self.context.print('    Adding a background a background electron '
                               f'density of {self.bg_density / Bohr**3.} Å^-3')
            n_sR = n_sR.copy()  # Make a copy in order not to modify gs
            n_sR += self.bg_density / n_sR.shape[0]

        return n_sR

    def get_gridrefinement(self, gd):
        if self.equivalent_real_space_grids(gd, self.gs.gd):
            gridrefinement = 1
        elif self.equivalent_real_space_grids(gd, self.gs.finegd):
            gridrefinement = 2
        else:
            raise ValueError('The supplied gd is neither compatible with the '
                             'coarse nor the fine real-space grid of the '
                             'underlying ground state')

        return gridrefinement


class LocalGridFTCalculator(LocalFTCalculator):

    def calculate(self, qpd, add_f):
        """Calculate f(G) directly from the all-electron density on the cubic
        real-space grid."""
        n_sR = self.get_all_electron_density(qpd.gd)
        f_G = self._calculate(qpd, n_sR, add_f)

        return f_G

    def _calculate(self, qpd, n_sR, add_f):
        """In-place calculation of the plane-wave components."""
        # Calculate f(r)
        gd = qpd.gd
        f_R = gd.zeros()
        add_f(gd, n_sR, f_R)

        # FFT to reciprocal space
        f_G = fft_from_grid(f_R, qpd)  # G = 1D grid of |G|^2/2 < ecut

        return f_G

    @timer('Calculate the all-electron density')
    def get_all_electron_density(self, gd):
        """Calculate the all-electron (spin-)density."""
        self.context.print('    Calculating the all-electron density')
        return self.get_electron_density(gd)


class LocalPAWFTCalculator(LocalFTCalculator):

    def __init__(self, gs, context, bg_density=None,
                 rshelmax=-1, rshewmin=None):
        super().__init__(gs, context, bg_density=bg_density)

        self.engine = LocalPAWFTEngine(self.context, rshelmax, rshewmin)

    def calculate(self, qpd, add_f):
        """Calculate f(G) with an expansion of f(r) in real spherical harmonics
        inside the augmentation spheres."""
        # Retrieve the pseudo (spin-)density on the real-space grid
        nt_sR = self.get_pseudo_density(qpd.gd)  # R = 3D real-space grid

        # Retrieve the pseudo and all-electron atomic centered densities inside
        # the augmentation spheres
        R_av, micro_setups = self.extract_atom_centered_quantities()

        # Let the engine perform the in-place calculation
        f_G = self.engine.calculate(qpd, nt_sR, R_av, micro_setups, add_f)

        return f_G

    def get_pseudo_density(self, gd):
        """Get the pseudo (spin-)density of the ground state."""
        return self.get_electron_density(gd, pseudo=True)

    def extract_atom_centered_quantities(self):
        """Extract all relevant atom centered quantities that the engine needs
        in order to calculate PAW corrections. Most of the information is
        bundled as a list of MicroSetups for each atom."""
        R_av = self.gs.atoms.positions / Bohr

        micro_setups = [extract_micro_setup(self.gs, a)
                        for a in range(len(self.gs.atoms))]

        return R_av, micro_setups


class MicroSetup:

    def __init__(self, rgd, Y_nL, n_sLg, nt_sLg):
        self.rgd = rgd
        self.Y_nL = Y_nL
        self.n_sLg = n_sLg
        self.nt_sLg = nt_sLg

    def evaluate_function(self, add_f):
        """Evaluate a given function f(r) on the angular and radial grids."""
        f_ng = np.array([self.rgd.zeros() for n in range(self.Y_nL.shape[0])])
        for n, Y_L in enumerate(self.Y_nL):
            n_sg = Y_L @ self.n_sLg
            add_f(self.rgd, n_sg, f_ng[n])
        return f_ng

    def evaluate_paw_correction(self, add_f):
        r"""Evaluate Δf_a[n_a,ñ_a](r) for a given function f(r).

        Returns
        -------
        df_ng : nd.array
            (f_ng - ft_ng) where (n=Lebedev index, g=radial grid index)
        """
        rgd = self.rgd
        f_g = rgd.zeros()
        ft_g = rgd.zeros()
        df_ng = np.array([rgd.zeros() for n in range(self.Y_nL.shape[0])])
        for n, Y_L in enumerate(self.Y_nL):
            f_g[:] = 0.
            n_sg = Y_L @ self.n_sLg
            add_f(rgd, n_sg, f_g)

            ft_g[:] = 0.
            nt_sg = Y_L @ self.nt_sLg
            add_f(rgd, nt_sg, ft_g)

            df_ng[n, :] = f_g - ft_g

        return df_ng


def extract_micro_setup(gs, a):
    """Extract the all-electron and pseudo spin-densities inside augmentation
    sphere a, as well as the relevant radial and angular grid data.

    Returns
    -------
    micro_setup : MicroSetup
    """
    pawdata = gs.pawdatasets[a]
    # Radial grid descriptor:
    rgd = pawdata.xc_correction.rgd
    # Spherical harmonics on the Lebedev quadrature:
    Y_nL = pawdata.xc_correction.Y_nL

    D_sp = gs.D_asp[a]  # atomic density matrix
    n_sLg, nt_sLg = calculate_atom_centered_densities(pawdata, D_sp)

    return MicroSetup(rgd, Y_nL, n_sLg, nt_sLg)


def calculate_atom_centered_densities(pawdata, D_sp):
    """Calculate the AE and pseudo densities inside the augmentation sphere.

    Returns
    -------
    n_sLg : nd.array
        all-electron density
    nt_sLg : nd.array
        pseudo density
    (s=spin, L=(l,m) spherical harmonic index, g=radial grid index)
    """
    n_qg = pawdata.xc_correction.n_qg
    nt_qg = pawdata.xc_correction.nt_qg
    nc_g = pawdata.xc_correction.nc_g
    nct_g = pawdata.xc_correction.nct_g

    B_pqL = pawdata.xc_correction.B_pqL
    D_sLq = np.inner(D_sp, B_pqL.T)
    nspins = len(D_sp)

    n_sLg = D_sLq @ n_qg
    nt_sLg = D_sLq @ nt_qg

    # Add core density
    n_sLg[:, 0] += np.sqrt(4. * np.pi) / nspins * nc_g
    nt_sLg[:, 0] += np.sqrt(4. * np.pi) / nspins * nct_g

    return n_sLg, nt_sLg


class LocalPAWFTEngine:

    def __init__(self, context, rshelmax=-1, rshewmin=None):
        """Construct the engine."""
        self.context = context
        self.rshelmax = rshelmax
        self.rshewmin = rshewmin

        self._add_f = None

    def calculate(self, qpd, nt_sR, R_av, micro_setups, add_f):
        r"""Calculate the Fourier transform f(G) by splitting up the
        calculation into a pseudo density contribution and a PAW correction
        accounting for the difference

        Δf[n,ñ](r) = f(n(r)) - f(ñ(r)),

        such that:

        f(G) = f[ñ](G) + Δf[n,ñ](G).

        See [PRB 103, 245110 (2021)] for definitions and notation details."""
        self._add_f = add_f

        ft_G = self.calculate_pseudo_contribution(qpd, nt_sR)
        fPAW_G = self.calculate_paw_corrections(qpd, R_av, micro_setups)

        return ft_G + fPAW_G

    def calculate_pseudo_contribution(self, qpd, nt_sR):
        """Calculate the pseudo density contribution by performing a FFT of
        f(ñ(r)) on the cubic real-space grid.

        NB: This operation assumes that the function f is a slowly varrying
        function of the pseudo density ñ(r) everywhere in space, such that
        f(ñ(r)) is accurately described on the cubic real-space grid."""
        # Calculate ft(r) (t=tilde=pseudo)
        gd = qpd.gd
        ft_R = gd.zeros()
        self._add_f(gd, nt_sR, ft_R)

        # FFT to reciprocal space
        ft_G = fft_from_grid(ft_R, qpd)  # G = 1D grid of |G|^2/2 < ecut

        return ft_G

    @timer('Calculate PAW corrections')
    def calculate_paw_corrections(self, qpd, R_av, micro_setups):
        r"""Calculate the PAW corrections to f(G), for each augmentation sphere
        at a time:
                      __
                      \   /
        Δf[n,ñ](G) =  /   |dr Δf_a[n_a,ñ_a](r - R_a) e^(-iG.r)
                      ‾‾  /
                      a    V0

        where Δf_a is the atom centered difference between the all electron
        and pseudo quantities inside augmentation sphere a:

        Δf_a[n_a,ñ_a](r) = f(n_a(r)) - f(ñ_a(r)).
        """
        self.context.print('    Calculating PAW corrections\n')

        # Extract reciprocal lattice vectors
        nG = qpd.ngmax
        G_Gv = qpd.get_reciprocal_vectors(add_q=False)
        assert G_Gv.shape[0] == nG

        # Allocate output array
        fPAW_G = np.zeros(nG, dtype=complex)

        # Distribute plane waves
        G_myG = self._distribute_correction(nG)
        G_myGv = G_Gv[G_myG]

        # Calculate and add the PAW corrections from each augmentation sphere
        for a, (R_v, micro_setup) in enumerate(zip(R_av, micro_setups)):
            self._add_paw_correction(a, R_v, micro_setup,
                                     G_myG, G_myGv, fPAW_G)

        self.context.comm.sum(fPAW_G)

        return fPAW_G

    def _distribute_correction(self, nG):
        comm = self.context.comm
        nGpr = (nG + comm.size - 1) // comm.size
        Ga = min(comm.rank * nGpr, nG)
        Gb = min(Ga + nGpr, nG)

        return range(Ga, Gb)

    def _add_paw_correction(self, a, R_v, micro_setup, G_myG, G_myGv, fPAW_G):
        r"""Calculate the PAW correction of augmentation sphere a,

                              /
        Δf_a(G) = e^(-iG.R_a) |dr Δf_a[n_a,ñ_a](r) e^(-iG.r),
                              /
                              V0

        by expanding both the atom centered correction and the plane wave in
        real spherical harmonics, see [PRB 103, 245110 (2021)]:

                                 l               a
                              __ __             R_c
                              \  \      l    ^  /                   a
        Δf_a(G) = e^(-iG.R_a) /  /  (-i)  Y (G) |4πr^2 dr j(|G|r) Δf (r)
                              ‾‾ ‾‾        lm   /          l        lm
                              l m=-l            0

        The calculated atomic correction is then added to the output array."""
        # Radial and angular grid information
        rgd, Y_nL = micro_setup.rgd, micro_setup.Y_nL

        # Calculate df on Lebedev quadrature (angular grid) and radial grid
        df_ng = self._calculate_df(micro_setup)

        # Reduce radial grid to where df is nonzero
        df_ng, r_g, dv_g = self.get_reduced_radial_grid(df_ng, rgd)

        # Expand correction in real spherical harmonics
        rshe, info_string = self.perform_rshe(rgd, df_ng, Y_nL)
        self.print_rshe_info(a, info_string)

        # Expand the plane waves in real spherical harmonics (and spherical
        # Bessel functions)
        (ii_MmyG,
         j_gMmyG,
         Y_MmyG) = self._expand_plane_waves(G_myGv, r_g, rshe.L_M, rshe.l_M)

        # Calculate the PAW correction as an integral over the radial grid
        # and rshe coefficients
        with self.context.timer('Integrate PAW correction'):
            angular_coef_MmyG = ii_MmyG * Y_MmyG
            # Radial integral, dv = 4πr^2
            df_gM = rshe.f_gM
            radial_coef_MmyG = np.tensordot(j_gMmyG * df_gM[..., np.newaxis],
                                            dv_g, axes=([0, 0]))
            # Angular integral (sum over l,m)
            atomic_corr_myG = np.sum(angular_coef_MmyG * radial_coef_MmyG,
                                     axis=0)

            position_prefactor_myG = np.exp(-1j * np.inner(G_myGv, R_v))

            # Add to output array
            fPAW_G[G_myG] += position_prefactor_myG * atomic_corr_myG

    @timer('Calculate PAW correction inside augmentation spheres')
    def _calculate_df(self, micro_setup):
        return micro_setup.evaluate_paw_correction(self._add_f)

    @staticmethod
    def get_reduced_radial_grid(df_ng, rgd):
        """Get the radial grid parameters where df_ng is nonzero.

        Returns
        -------
        df_ng : nd.array
            df_ng on reduced radial grid
        r_g : nd.array
            radius of each point on the reduced radial grid
        dv_g : nd.array
            volume element of each point on the reduced radial grid
        """
        # Find PAW correction range
        dfmask_g = np.where(np.any(np.abs(df_ng) > 0., axis=0))
        ng = np.max(dfmask_g) + 1

        # Integrate only r-values inside PAW correction range
        df_ng = df_ng[:, :ng]
        r_g = rgd.r_g[:ng]
        dv_g = rgd.dv_g[:ng]

        return df_ng, r_g, dv_g

    @timer('Expand PAW correction in real spherical harmonics')
    def perform_rshe(self, rgd, df_ng, Y_nL):
        r"""Expand Δf_a[n_a,ñ_a](r) in real spherical harmonics."""
        return calculate_reduced_rshe(
            rgd, df_ng, Y_nL, self.rshelmax, self.rshewmin)

    def print_rshe_info(self, a, info_string):
        """Print information about the expansion at atom a."""
        info_string = f'    RSHE of atom {a}\n' + info_string
        self.context.print(
            info_string.replace('\n', '\n      ') + '\n')

    @timer('Expand plane waves in real spherical harmonics')
    def _expand_plane_waves(self, G_myGv, r_g, L_M, l_M):
        r"""Expand plane waves in spherical Bessel functions and real spherical
        harmonics:
                        l
                    __  __
         -iG.r      \   \      l             ^     ^
        e      = 4π /   /  (-i)  j (|G|r) Y (G) Y (r)
                    ‾‾  ‾‾        l        lm    lm
                    l  m=-l

        Returns
        -------
        ii_MmyG : nd.array
            (-i)^l for used (l,m) coefficients M
        j_gMmyG : nd.array
            j_l(|G|r) for used (l,m) coefficients M
        Y_MmyG : nd.array
                 ^
            Y_lm(K) for used (l,m) coefficients M
        """
        nmyG = G_myGv.shape[0]
        Gnorm_myG, Gdir_myGv = self._calculate_norm_and_direction(G_myGv)

        # Setup arrays to fully vectorize computations
        nM = len(L_M)
        (r_gMmyG, l_gMmyG,
         Gnorm_gMmyG) = [a.reshape(len(r_g), nM, nmyG)
                         for a in np.meshgrid(r_g, l_M, Gnorm_myG,
                                              indexing='ij')]

        with self.context.timer('Compute spherical bessel functions'):
            # Slow step
            j_gMmyG = spherical_jn(l_gMmyG, Gnorm_gMmyG * r_gMmyG)

        Y_MmyG = Yarr(L_M, Gdir_myGv)
        ii_MmyG = (-1j) ** np.repeat(l_M, nmyG).reshape((nM, nmyG))

        return ii_MmyG, j_gMmyG, Y_MmyG

    @staticmethod
    def _calculate_norm_and_direction(G_myGv):
        """Calculate the length and direction of reciprocal lattice vectors."""
        Gnorm_myG = np.linalg.norm(G_myGv, axis=1)
        Gdir_myGv = np.zeros_like(G_myGv)
        mask0 = np.where(Gnorm_myG != 0.)
        Gdir_myGv[mask0] = G_myGv[mask0] / Gnorm_myG[mask0][:, np.newaxis]

        return Gnorm_myG, Gdir_myGv


def fft_from_grid(f_R, qpd):
    r"""Perform a FFT to reciprocal space:
                                    __
           /                    V0  \
    f(G) = |dr f(r) e^(-iG.r) ≃ ‾‾  /  f(r) e^(-iG.r)
           /                    N   ‾‾
           V0                       r

    where N is the number of grid points."""
    Q_G = qpd.Q_qG[0]

    # Perform the FFT
    N = np.prod(qpd.gd.N_c)
    f_Q123 = qpd.gd.volume / N * np.fft.fftn(f_R)  # Q123 = 3D grid in Q-rep

    # Change the view of the plane-wave components from the 3D grid in the
    # Q-representation that numpy spits out to the 1D grid in the
    # G-representation, that GPAW relies on internally
    f_G = f_Q123.ravel()[Q_G]

    return f_G


# ---------- Local functions of the (spin-)density ---------- #


def add_total_density(gd, n_sR, n_R):
    n_R += np.sum(n_sR, axis=0)


def add_spin_polarization(gd, n_sR, nz_R):
    nz_R += calculate_spin_polarization(n_sR)


def calculate_spin_polarization(n_sR):
    return n_sR[0] - n_sR[1]


def add_LSDA_Wxc(gd, n_sR, Wxc_R):
    Wxc_R += calculate_LSDA_Wxc(gd, n_sR)


def calculate_LSDA_Wxc(gd, n_sR):
    """Calculate W_xc^z in the local spin-density approximation.

    For a collinear system:

                δE_xc[n,n^z]   1
    W_xc^z(r) = ‾‾‾‾‾‾‾‾‾‾‾‾ = ‾ [V_LSDA^↑(r) - V_LSDA^↓(r)]
                   δn^z(r)     2
    """
    # Allocate an array for the spin-dependent xc potential on the real
    # space grid
    v_sR = np.zeros(np.shape(n_sR))

    # Calculate the spin-dependent potential
    xc = XC('LDA')
    xc.calculate(gd, n_sR, v_sg=v_sR)

    return (v_sR[0] - v_sR[1]) / 2


def add_LSDA_spin_splitting(gd, n_sR, dxc_R):
    """Calculate and add the LSDA spin splitting to the output array.

    The spin splitting is defined as:

    Δ^(xc)(r) = - 2 B^(xc)(r) m(r) = - 2 W_xc^z n^z(r).
    """
    dxc_R += - 2. * calculate_LSDA_Wxc(gd, n_sR) \
        * calculate_spin_polarization(n_sR)


def add_LDA_dens_fxc(gd, n_sR, fxc_R, *, fxc):
    r"""Calculate the LDA density kernel and add it to the output array fxc_R.

    The LDA density kernel is given by:

                    ∂^2[ϵ_xc(n,m)n] |
    f_LDA^(00)(r) = ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾ |
                         ∂n^2       |n=n(r),m=m(r)
    """
    assert len(n_sR) == 1, \
        'The density kernel is untested for spin-polarized systems'

    if fxc == 'ALDA_x':
        fxc_R += -1. / 3. * (3. / np.pi)**(1. / 3.) * n_sR[0]**(-2. / 3.)
    else:
        assert fxc in ['ALDA_X', 'ALDA']
        kernel = LibXC(fxc[1:])
        fxc_sR = np.zeros_like(n_sR)
        kernel.xc.calculate_fxc_spinpaired(n_sR.ravel(), fxc_sR)

        fxc_R += fxc_sR[0]


def add_LSDA_trans_fxc(gd, n_sR, fxc_R, *, fxc):
    r"""Calculate the transverse LDA kernel and add it to the output arr. fxc_R

    The transverse LDA kernel is given by:

                     2  ∂[ϵ_xc(n,n^z)n] |
    f_LDA^(+-)(r) = ‾‾‾ ‾‾‾‾‾‾‾‾‾‾‾‾‾‾  |
                    n^z      ∂n^z       |n=n(r),n^z=n^z(r)

                    V_LSDA^↑(r) - V_LSDA^↓(r)
                  = ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                              n^z(r)
    """
    assert len(n_sR) == 2  # nspins
    nz_R = n_sR[0] - n_sR[1]

    if fxc == 'ALDA_x':
        fxc_R += - (6. / np.pi)**(1. / 3.) \
            * (n_sR[0]**(1. / 3.) - n_sR[1]**(1. / 3.)) / nz_R
    else:
        assert fxc in ['ALDA_X', 'ALDA']
        v_sR = np.zeros(np.shape(n_sR))
        xc = XC(fxc[1:])
        xc.calculate(gd, n_sR, v_sg=v_sR)

        fxc_R += (v_sR[0] - v_sR[1]) / nz_R



"""Test for elph/gmatrix"""
import numpy as np
import pytest

from ase.build import bulk
# from ase.phonons import Phonons

from gpaw import GPAW
from gpaw.elph import ElectronPhononMatrix


@pytest.mark.serial
@pytest.mark.elph
def test_gmatrix(module_tmp_path, supercell_cache):
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True)
    supercell_cache
    elph = ElectronPhononMatrix(atoms, 'supercell', 'elph')
    q = [[0, 0, 0], [0.5, 0.5, 0.5]]

    # phonon = Phonons(atoms, name='elph', supercell=(2,1,1),
    #                  center_refcell=True)
    # phonon.read()
    # w_ql = phonon.band_structure(q, modes=False)
    # print(w_ql)
    # [[0.00124087  0.00044091  0.00132004  0.0421112   0.04212737  0.04218485]
    # [0.03031018  0.03031948  0.03041029  0.03041035  0.04326759  0.04327498]]

    calc = GPAW(mode='lcao',
                basis='sz(dzp)',
                kpts={'size': (2, 2, 2), 'gamma': False},
                symmetry='off',
                txt='li_gs_nosym.txt')
    atoms.calc = calc
    atoms.get_potential_energy()

    g_sqklnn = elph.bloch_matrix(calc, k_qc=q,
                                 savetofile=False, prefactor=False)

    assert g_sqklnn.shape == (1, 2, 8, 6, 2, 2)

    # NOTE: It seems g is Hermitian if q=0 and symmetric otherwise. CHECK THIS!

    # Li has lots of degenerate phonon modes, average/sum those

    # q = 0
    # accoustic sum rule
    assert np.allclose(g_sqklnn[0, 0, :, 0:3], 0.)

    g_knn = np.sum(g_sqklnn[0, 0], axis=1, initial=3)  # modes 4-6
    assert g_knn.shape == (8, 2, 2)
    # Hermitian
    assert np.allclose(g_knn[:, 0, 1], g_knn[:, 1, 0].conj())
    # and check one specific value
    print(g_knn[0, 0, 1])

    assert g_knn[0, 0, 1].real == pytest.approx(2.9999902, rel=1e-4)
    assert (g_knn[0, 0, 1].imag == pytest.approx(0.1847967, abs=0.004) or
            g_knn[0, 0, 1].imag == pytest.approx(-0.1624835, abs=0.004))
    # q = 1
    g_knn = np.sum(g_sqklnn[0, 1], axis=1, initial=4)  # modes 5-6
    assert g_knn.shape == (8, 2, 2)
    # Hermitian, actually symmetric
    assert np.allclose(g_knn[:, 0, 1], g_knn[:, 1, 0])
    # and check one specific value
    # print(g_knn)
    print(g_knn[0, 0, 1])
    assert g_knn[0, 0, 1] == pytest.approx(3.510762 - 0.238049j, abs=0.01)



# Test of the handling of degenerate bands in response code
import pytest
from ase import Atoms
from gpaw import GPAW, PW
from gpaw.response.df import DielectricFunction
from gpaw.test import findpeak, equal
import numpy as np


def get_hydrogen_chain_dielectric_function(NH, NK):
    a = Atoms('H', cell=[1, 1, 2], pbc=True)
    a.center()
    a = a.repeat((1, 1, NH))
    a.calc = GPAW(mode=PW(200, force_complex_dtype=True),
                  kpts={'size': (1, 1, NK), 'gamma': True},
                  parallel={'band': 1},
                  gpts=(10, 10, 10 * NH))
    a.get_potential_energy()
    a.calc.diagonalize_full_hamiltonian(nbands=2 * NH)
    a.calc.write('H_chain.gpw', 'all')

    DF = DielectricFunction('H_chain.gpw', ecut=1e-3, hilbert=False,
                            frequencies={'type': 'nonlinear',
                                         'domega0': None,
                                         'omega2': np.inf,
                                         'omegamax': None},
                            intraband=False)
    eps_NLF, eps_LF = DF.get_dielectric_function(direction='z')
    omega_w = DF.get_frequencies()
    return omega_w, eps_LF


@pytest.mark.serial
@pytest.mark.response
def test_hyd_chain_response(in_tmp_dir):
    NH_i = [2**n for n in [0, 2]]
    NK_i = [2**n for n in [4, 2]]

    opeak_old = np.nan
    peak_old = np.nan

    for NH, NK in zip(NH_i, NK_i):
        omega_w, eps_w = get_hydrogen_chain_dielectric_function(NH, NK)
        eels_w = -(1. / eps_w).imag
        opeak, peak = findpeak(omega_w, eels_w)

        # Test for consistency
        if not np.isnan(opeak_old):
            equal(opeak, opeak_old, tolerance=1e-3)
            equal(peak, peak_old, tolerance=1e-3)
        opeak_old = opeak
        peak_old = peak



# General modules
import pytest
import numpy as np

# Script modules
from gpaw import GPAW
from gpaw.test import findpeak
from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.susceptibility import (ChiFactory, spectral_decomposition,
                                          EigendecomposedSpectrum,
                                          read_full_spectral_weight,
                                          read_eigenmode_lineshapes)
from gpaw.response.fxc_kernels import AdiabaticFXCCalculator
from gpaw.response.dyson import HXCScaling
from gpaw.response.pair_functions import read_susceptibility_array
from gpaw.test.conftest import response_band_cutoff


@pytest.mark.kspair
@pytest.mark.response
def test_response_cobalt_sf_gssALDA(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    fxc = 'ALDA'
    q_qc = [[0.0, 0.0, 0.0], [1. / 4., 0.0, 0.0]]  # Two q-points along G-M
    # We make sure to have exactly 49 frequency points, so that one rank will
    # have no block distributed frequencies when world.size == 8
    frq_w = np.linspace(-0.6, 1.2, 49)
    eta = 0.2

    rshelmax = 0
    hxc_scaling = HXCScaling('fm')
    ecut = 250
    reduced_ecut = 100  # ecut for eigenmode analysis
    pos_eigs = 2  # majority modes
    neg_eigs = 0  # minority modes
    nblocks = 'max'

    # ---------- Script ---------- #

    # Initialize objects to calculat Chi
    context = ResponseContext()
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['co_pw']
    gs = ResponseGroundStateAdapter(calc)
    chiks_calc = ChiKSCalculator(gs, context,
                                 nbands=nbands,
                                 ecut=ecut,
                                 gammacentered=True,
                                 nblocks=nblocks)
    fxc_calculator = AdiabaticFXCCalculator.from_rshe_parameters(
        gs, context, rshelmax=rshelmax)
    chi_factory = ChiFactory(chiks_calc, fxc_calculator)

    for q, q_c in enumerate(q_qc):
        complex_frequencies = frq_w + 1.j * eta
        chiks, chi = chi_factory('+-', q_c, complex_frequencies,
                                 fxc=fxc, hxc_scaling=hxc_scaling)

        # Check that the dissipative part of the susceptibility can be
        # reconstructed from the EigendecomposedSpectrum
        chid = chi.copy_dissipative_part()
        spectrum = EigendecomposedSpectrum.from_chi(chi)
        assert np.allclose(-np.pi * spectrum.A_wGG, chid.array)

        # Perform a spectral decomposition of the magnetic excitations in the
        # Kohn-Sham and many-body susceptibilities, writing the full spectral
        # weight of both the majority and minority spectra
        Aksmaj, Aksmin = spectral_decomposition(chiks)
        Aksmaj.write_full_spectral_weight(f'Aksmaj_q{q}.dat')
        Aksmin.write_full_spectral_weight(f'Aksmin_q{q}.dat')
        Amaj, Amin = spectral_decomposition(chi,
                                            pos_eigs=pos_eigs,
                                            neg_eigs=neg_eigs)
        Amaj.write_full_spectral_weight(f'Amaj_q{q}.dat')
        Amin.write_full_spectral_weight(f'Amin_q{q}.dat')
        # Write also the majority magnon modes of the many-body system
        Amaj.write_eigenmode_lineshapes(f'Amaj_modes_q{q}.dat', nmodes=2)

        # Perform an analogous spectral decomposition of the many-body
        # susceptibility in a reduced plane-wave basis and write the decomposed
        # spectrum to a file
        chi = chi.copy_with_reduced_ecut(reduced_ecut)
        chi.write_diagonal(f'chiwG_q{q}.pckl')
        Amaj, Amin = spectral_decomposition(chi,
                                            pos_eigs=pos_eigs,
                                            neg_eigs=neg_eigs)
        Amaj.write(f'Amaj_q{q}.pckl')
        Amin.write(f'Amin_q{q}.pckl')
        assert f'{fxc},+-' in chi_factory.fxc_kernel_cache

    context.write_timer()
    context.comm.barrier()

    # Read data
    wksmaj0_w, Aksmaj0_w = read_full_spectral_weight('Aksmaj_q0.dat')
    wksmin0_w, Aksmin0_w = read_full_spectral_weight('Aksmin_q0.dat')
    wksmaj1_w, Aksmaj1_w = read_full_spectral_weight('Aksmaj_q1.dat')
    wksmin1_w, Aksmin1_w = read_full_spectral_weight('Aksmin_q1.dat')
    wmaj0_w, Amaj0_w = read_full_spectral_weight('Amaj_q0.dat')
    wmin0_w, Amin0_w = read_full_spectral_weight('Amin_q0.dat')
    wmaj1_w, Amaj1_w = read_full_spectral_weight('Amaj_q1.dat')
    wmin1_w, Amin1_w = read_full_spectral_weight('Amin_q1.dat')
    wa0_w, a0_wm = read_eigenmode_lineshapes('Amaj_modes_q0.dat')
    wa1_w, a1_wm = read_eigenmode_lineshapes('Amaj_modes_q1.dat')
    w0_w, _, chi0_wG = read_susceptibility_array('chiwG_q0.pckl')
    w1_w, _, chi1_wG = read_susceptibility_array('chiwG_q1.pckl')
    Amaj0 = EigendecomposedSpectrum.from_file('Amaj_q0.pckl')
    Amaj1 = EigendecomposedSpectrum.from_file('Amaj_q1.pckl')

    # Find acoustic magnon mode
    wpeak00, Ipeak00 = findpeak(w0_w, -chi0_wG[:, 0].imag)
    wpeak01, Ipeak01 = findpeak(w1_w, -chi1_wG[:, 0].imag)
    # Find optical magnon mode
    wpeak10, Ipeak10 = findpeak(w0_w, -chi0_wG[:, 1].imag)
    wpeak11, Ipeak11 = findpeak(w1_w, -chi1_wG[:, 1].imag)

    # Extract the eigenmodes from the eigendecomposed spectrum in the reduced
    # plane-wave basis
    ar0_wm = Amaj0.get_eigenmode_lineshapes(nmodes=pos_eigs)
    ar1_wm = Amaj1.get_eigenmode_lineshapes(nmodes=pos_eigs)

    # Find peaks in eigenmodes (in full and reduced basis)
    mpeak00, Speak00 = findpeak(wa0_w, a0_wm[:, 0])
    mpeak01, Speak01 = findpeak(wa1_w, a1_wm[:, 0])
    mpeak10, Speak10 = findpeak(wa0_w, a0_wm[:, 1])
    mpeak11, Speak11 = findpeak(wa1_w, a1_wm[:, 1])
    mrpeak00, Srpeak00 = findpeak(Amaj0.omega_w, ar0_wm[:, 0])
    mrpeak01, Srpeak01 = findpeak(Amaj1.omega_w, ar1_wm[:, 0])
    mrpeak10, Srpeak10 = findpeak(Amaj0.omega_w, ar0_wm[:, 1])
    mrpeak11, Srpeak11 = findpeak(Amaj1.omega_w, ar1_wm[:, 1])

    # Calculate the majority spectral enhancement at the acoustic magnon maxima
    w0 = np.argmin(np.abs(Amaj0.omega_w - wpeak00))
    w1 = np.argmin(np.abs(Amaj1.omega_w - wpeak01))
    enh0 = Amaj0_w[w0] / Aksmaj0_w[w0]
    enh1 = Amaj1_w[w1] / Aksmaj1_w[w1]

    # Calculate the minority spectral enhancement at 600 meV (corresponding to
    # -600 meV on original frequency grid)
    min_enh0 = Amin0_w[0] / Aksmin0_w[0]
    min_enh1 = Amin1_w[0] / Aksmin1_w[0]

    if context.comm.rank == 0:
        # import matplotlib.pyplot as plt
        # # Plot the magnon lineshapes
        # # q=0
        # plt.subplot(2, 3, 1)
        # plt.plot(w0_w, -chi0_wG[:, 0].imag)
        # plt.axvline(wpeak00, c='0.5', linewidth=0.8)
        # plt.plot(w0_w, -chi0_wG[:, 1].imag)
        # plt.axvline(wpeak10, c='0.5', linewidth=0.8)
        # plt.plot(Amaj0.omega_w, Amaj0.s_we[:, 0])
        # plt.plot(Amaj0.omega_w, Amaj0.s_we[:, 1])
        # plt.plot(wa0_w, a0_wm[:, 0])
        # plt.plot(wa0_w, a0_wm[:, 1])
        # plt.plot(Amaj0.omega_w, ar0_wm[:, 0])
        # plt.plot(Amaj0.omega_w, ar0_wm[:, 1])
        # # q=1
        # plt.subplot(2, 3, 4)
        # plt.plot(w1_w, -chi1_wG[:, 0].imag)
        # plt.axvline(wpeak01, c='0.5', linewidth=0.8)
        # plt.plot(w1_w, -chi1_wG[:, 1].imag)
        # plt.axvline(wpeak11, c='0.5', linewidth=0.8)
        # plt.plot(Amaj1.omega_w, Amaj1.s_we[:, 0])
        # plt.plot(Amaj1.omega_w, Amaj1.s_we[:, 1])
        # plt.plot(wa1_w, a1_wm[:, 0])
        # plt.plot(wa1_w, a1_wm[:, 1])
        # plt.plot(Amaj1.omega_w, ar1_wm[:, 0])
        # plt.plot(Amaj1.omega_w, ar1_wm[:, 1])
        # # Plot full spectral weight of majority excitations
        # # q=0
        # plt.subplot(2, 3, 2)
        # plt.plot(wmaj0_w, Amaj0_w)
        # plt.plot(wksmaj0_w, Aksmaj0_w)
        # plt.plot(Amaj0.omega_w, Amaj0.A_w)
        # plt.axvline(wpeak00, c='0.5', linewidth=0.8)
        # # q=1
        # plt.subplot(2, 3, 5)
        # plt.plot(wmaj1_w, Amaj1_w)
        # plt.plot(wksmaj1_w, Aksmaj1_w)
        # plt.plot(Amaj1.omega_w, Amaj1.A_w)
        # plt.axvline(wpeak01, c='0.5', linewidth=0.8)
        # # Plot full spectral weight of minority excitations
        # # q=0
        # plt.subplot(2, 3, 3)
        # plt.plot(wmin0_w, Amin0_w)
        # plt.plot(wksmin0_w, Aksmin0_w)
        # # q=1
        # plt.subplot(2, 3, 6)
        # plt.plot(wmin1_w, Amin1_w)
        # plt.plot(wksmin1_w, Aksmin1_w)
        # plt.show()

        # Print values
        print(hxc_scaling.lambd)
        print(wpeak00, wpeak01, wpeak10, wpeak11)
        print(Ipeak00, Ipeak01, Ipeak10, Ipeak11)
        print(mpeak00, mpeak01, mpeak10, mpeak11)
        print(Speak00, Speak01, Speak10, Speak11)
        print(mrpeak00, mrpeak01, mrpeak10, mrpeak11)
        print(Srpeak00, Srpeak01, Srpeak10, Srpeak11)
        print(enh0, enh1, min_enh0, min_enh1)

    # Test kernel scaling
    assert hxc_scaling.lambd == pytest.approx(0.9675, abs=0.005)

    # Test magnon frequencies
    assert wpeak00 == pytest.approx(-0.0281, abs=0.005)
    assert wpeak01 == pytest.approx(0.218, abs=0.01)
    assert wpeak10 == pytest.approx(0.508, abs=0.01)
    assert wpeak11 == pytest.approx(0.637, abs=0.01)

    # Test magnon amplitudes
    assert Ipeak00 == pytest.approx(2.897, abs=0.01)
    assert Ipeak01 == pytest.approx(2.245, abs=0.01)
    assert Ipeak10 == pytest.approx(1.090, abs=0.01)
    assert Ipeak11 == pytest.approx(1.023, abs=0.01)

    # Test magnon frequency consistency
    assert mpeak00 == pytest.approx(wpeak00, abs=0.005)
    assert mpeak01 == pytest.approx(wpeak01, abs=0.01)
    assert mpeak10 == pytest.approx(wpeak10, abs=0.01)
    assert mpeak11 == pytest.approx(wpeak11, abs=0.01)
    assert mrpeak00 == pytest.approx(wpeak00, abs=0.005)
    assert mrpeak01 == pytest.approx(wpeak01, abs=0.01)
    assert mrpeak10 == pytest.approx(wpeak10, abs=0.01)
    assert mrpeak11 == pytest.approx(wpeak11, abs=0.01)

    # Test magnon mode eigenvalues at extrema
    assert Speak00 == pytest.approx(8.409, abs=0.02)
    assert Speak01 == pytest.approx(6.734, abs=0.02)
    assert Speak10 == pytest.approx(3.800, abs=0.02)
    assert Speak11 == pytest.approx(3.683, abs=0.02)
    assert Srpeak00 == pytest.approx(6.402, abs=0.02)
    assert Srpeak01 == pytest.approx(5.087, abs=0.02)
    assert Srpeak10 == pytest.approx(2.837, abs=0.02)
    assert Srpeak11 == pytest.approx(2.692, abs=0.02)

    # Test enhancement factors
    assert enh0 == pytest.approx(36.77, abs=0.1)
    assert enh1 == pytest.approx(24.10, abs=0.1)
    assert min_enh0 == pytest.approx(1.141, abs=0.01)
    assert min_enh1 == pytest.approx(1.162, abs=0.01)



import numpy as np
import pytest

from ase.build import fcc111

from gpaw import GPAW
from gpaw.mpi import world, serial_comm
from gpaw.lcaotddft.wfwriter import WaveFunctionReader

from gpaw.test import only_on_master
from . import (parallel_options, calculate_error, calculate_time_propagation,
               check_wfs)

pytestmark = pytest.mark.usefixtures('module_tmp_path')

parallel_i = parallel_options(include_kpt=True)


@pytest.mark.rttddft
@pytest.fixture(scope='module')
@only_on_master(world)
def initialize_system():
    comm = serial_comm

    # Ground-state calculation
    atoms = fcc111('Al', size=(1, 1, 2), vacuum=4.0)
    atoms.symbols[0] = 'Li'
    calc = GPAW(nbands=4,
                h=0.4,
                kpts={'size': (3, 3, 1)},
                basis='sz(dzp)',
                mode='lcao',
                convergence={'density': 1e-8},
                symmetry={'point_group': False},
                communicator=comm,
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    calculate_time_propagation('gs.gpw',
                               kick=[0, 0, 1e-5],
                               communicator=comm)


@pytest.mark.rttddft
def test_propagated_wave_function(initialize_system, module_tmp_path):
    wfr = WaveFunctionReader(module_tmp_path / 'wf.ulm')
    coeff = wfr[-1].wave_functions.coefficients
    coeff = coeff[np.ix_([0], [0, 1], [1, 3], [0, 1, 2])]
    # Normalize the wave function sign
    coeff = np.sign(coeff.real[..., 0, np.newaxis]) * coeff
    ref = [[[[5.4119034398864430e-01 + 4.6958807325576735e-01j,
              -5.8836045927143954e-01 - 5.1047688429408378e-01j,
              -6.5609314466400698e-06 - 5.8109609173527947e-06j],
             [1.6425837099429430e-06 - 1.4779657236004961e-06j,
              -8.7230715222772428e-07 + 8.9374679369814926e-07j,
              3.1300283337601806e+00 - 2.7306795126551076e+00j]],
            [[1.9820345503468246e+00 + 1.0562314330323577e+00j,
              -1.5008623926242098e-01 + 4.5817475674967340e-01j,
              -4.8385783015916195e-01 - 5.3676335879786385e-01j],
             [2.4227856141643818e+00 + 3.7767002050641824e-01j,
              -2.6174901880264838e+00 + 1.9885717875694848e+00j,
              7.2641847473298660e-01 + 1.6020733667409095e+00j]]]]
    err = calculate_error(coeff, ref)
    assert err < 7e-9


@pytest.mark.rttddft
@pytest.mark.parametrize('parallel', parallel_i)
def test_propagation(initialize_system, module_tmp_path, parallel, in_tmp_dir):
    calculate_time_propagation(module_tmp_path / 'gs.gpw',
                               kick=[0, 0, 1e-5],
                               parallel=parallel)
    check_wfs(module_tmp_path / 'wf.ulm', 'wf.ulm', atol=1e-12)



import pytest
from ase.build import molecule
from gpaw import GPAW


@pytest.mark.libxc
@pytest.mark.gllb
def test_gllbghost():
    atoms = molecule('H2')
    atoms.center(vacuum=2)
    calc = GPAW(mode='lcao', basis='dzp', setups={0: 'paw', 1: 'ghost'},
                xc='GLLBSC')
    atoms.calc = calc
    atoms.get_potential_energy()



import pytest
from ase import Atoms
from gpaw import GPAW, FermiDirac
from gpaw.mixer import MixerSum


def test_fixmom():
    a = 2.87
    bulk = Atoms('Fe2',
                 scaled_positions=[(0, 0, 0), (0.5, 0.5, 0.5)],
                 magmoms=[2.20, 2.20],
                 cell=(a, a, a),
                 pbc=True)
    mom0 = sum(bulk.get_initial_magnetic_moments())
    conv = {'eigenstates': 0.1, 'density': 0.001, 'energy': 0.01}
    calc = GPAW(mode='pw',
                mixer=MixerSum(0.1, 3),
                nbands=11,
                kpts=(3, 3, 3),
                convergence=conv,
                occupations=FermiDirac(0.1, fixmagmom=True))
    bulk.calc = calc
    bulk.get_potential_energy()
    mom = bulk.get_magnetic_moment()
    assert mom == pytest.approx(mom0, abs=0.005)



import numpy as np
import pytest
from ase import Atoms

from gpaw.grid_descriptor import GridDescriptor
from gpaw.hybrids.kpts import PWKPoint
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.mpi import world
from gpaw.projections import Projections
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC
from gpaw.hybrids.wstc import WignerSeitzTruncatedCoulomb as WSTC
from gpaw.spline import Spline
from gpaw.symmetry import Symmetry
from gpaw.wavefunctions.arrays import PlaneWaveExpansionWaveFunctions


class AP:
    my_indices = [0]
    comm = world
    rank_a = [0]


r2 = np.linspace(0, 1, 51)**2


class Setup:
    Delta_iiL = np.zeros((1, 1, 1)) + 0.1
    X_p = np.zeros(1) + 0.3
    ExxC = -10.0
    ghat_l = [Spline(0, 1.0, 1 - r2 * (1 - 2 * r2))]


@pytest.mark.xfail
def test_exx_derivs():
    if world.size > 1:
        from unittest import SkipTest
        raise SkipTest

    N = 20
    L = 2.5
    nb = 2
    spos_ac = np.zeros((1, 3)) + 0.25

    gd = GridDescriptor([N, N, N], np.eye(3) * L)
    sym = Symmetry([], gd.cell_cv)
    kd = KPointDescriptor(None)
    kd.set_symmetry(Atoms(pbc=True), sym)
    coulomb = WSTC(gd.cell_cv, kd.N_c)
    pd = PWDescriptor(10, gd, complex, kd)

    data = pd.zeros(nb)
    data[0, 1] = 3.0
    data[1, 2] = -2.5
    psit = PlaneWaveExpansionWaveFunctions(nb, pd, data=data)

    proj = Projections(nb, [1], AP(), world, dtype=complex)

    pt = PWLFC([[Spline(0, 1.0, 1 - r2 * (1 - 2 * r2))]], pd)
    pt.set_positions(spos_ac)

    f_n = np.array([1.0, 0.5])
    kpt = PWKPoint(psit, proj, f_n, np.array([0.0, 0.0, 0.0]), 1.0)

    # xx = EXX(kd, [Setup()], pt, coulomb, spos_ac)
    xx = (kd, [Setup()], pt, coulomb, spos_ac)

    psit.matrix_elements(pt, out=proj)
    C = 0.79
    VV_aii = {a: np.einsum('n, ni, nj -> ij', f_n, P_ni, P_ni.conj()) * C
              for a, P_ni in proj.items()}

    x = xx.calculate([kpt], [kpt], VV_aii, derivatives=True)
    v = x[3][0][0, 1]

    eps = 0.00001

    data[0, 1] = 3 + eps
    psit.matrix_elements(pt, out=proj)
    VV_aii = {a: np.einsum('n, ni, nj -> ij', f_n, P_ni, P_ni.conj()) * C
              for a, P_ni in proj.items()}
    xp = xx.calculate([kpt], [kpt], VV_aii)

    data[0, 1] = 3 - eps
    psit.matrix_elements(pt, out=proj)
    VV_aii = {a: np.einsum('n, ni, nj -> ij', f_n, P_ni, P_ni.conj()) * C
              for a, P_ni in proj.items()}
    xm = xx.calculate([kpt], [kpt], VV_aii)

    d = (xp[0] + xp[1] - xm[0] - xm[1]) / (2 * eps) * N**6 / L**3 / 2
    assert abs(v - d) < 1e-10, (v, d)



"""
Calculates Raman matrices.

i -> j -> m -> n
i, n are valence; j, m are conduction, also i=n in the end
see https://doi.org/10.1038/s41467-020-16529-6
"""
import numpy as np

from ase.units import invcm, Hartree
from ase.utils.filecache import MultiFileJSONCache
from gpaw.calculator import GPAW
from gpaw.typing import ArrayND

# TODO: only take kd, don't need whole calculator


class ResonantRamanCalculator:
    def __init__(
        self,
        calc: GPAW,
        wph_w: ArrayND,
        momname: str = "mom_skvnm.npy",
        elphname: str = "gsqklnn.npy",
        raman_name: str = "Rlab",
    ):
        """Resonant Raman Matrix calculator

        Parameters
        ----------
        calc: GPAW
            GPAW calculator object.
        w_ph: str, np.ndarray
            Zone centre phonon frequencies in eV
        momname: str
            Name of momentum file
        elphname: str
            Name of electron-phonon file
        raman_name: str
            Name of Rlab file cache. Default 'Rlab'
        """
        # those won't make sense here
        assert calc.wfs.gd.comm.size == 1
        assert calc.wfs.bd.comm.size == 1
        self.kd = calc.wfs.kd
        self.calc = calc

        # Phonon frequencies
        if isinstance(wph_w, str):
            wph_w = np.load(wph_w)
        assert max(wph_w) < 1.0  # else not eV units
        self.w_ph = wph_w

        # Load files
        mom_skvnm = np.load(momname, mmap_mode="c")
        g_sqklnn = np.load(elphname, mmap_mode="c")  # [s,q=0,k,l,n,m]
        self.mom_skvnm = mom_skvnm
        self.g_sqklnn = g_sqklnn

        # Define a few more variables
        nspins = g_sqklnn.shape[0]
        nk = g_sqklnn.shape[2]
        assert wph_w.shape[0] == g_sqklnn.shape[3]
        assert mom_skvnm.shape[0] == nspins
        assert mom_skvnm.shape[1] == nk
        assert mom_skvnm.shape[-1] == g_sqklnn.shape[-1]

        # JSON cache
        self.raman_cache = MultiFileJSONCache(raman_name)
        with self.raman_cache.lock("phonon_frequencies") as handle:
            if handle is not None:
                handle.save(wph_w)
        # Resonant part of Raman tensor does not have a frequency grid
        with self.raman_cache.lock("frequency_grid") as handle:
            if handle is not None:
                handle.save(None)

    @classmethod
    def resonant_term(
        cls,
        f_vc: ArrayND,
        E_vc: ArrayND,
        mom_dnn: ArrayND,
        elph_lnn: ArrayND,
        nc: int,
        nv: int,
        w_in: float,
        wph_w: ArrayND,
    ) -> ArrayND:
        """Resonant term of the Raman tensor"""
        nmodes = elph_lnn.shape[0]
        term_l = np.zeros((nmodes), dtype=complex)
        t_ij = f_vc * mom_dnn[0, :nv, nc:] / (w_in - E_vc)
        for l in range(nmodes):
            t_xx = elph_lnn[l]
            t_mn = f_vc.T * mom_dnn[1, nc:, :nv] / (w_in - wph_w[l] - E_vc.T)
            term_l[l] += np.einsum("sj,jm,ms", t_ij, t_xx[nc:, nc:], t_mn)
            term_l[l] -= np.einsum("is,ni,sn", t_ij, t_xx[:nv, :nv], t_mn)
        return term_l

    def calculate(self, w_in, d_i, d_o, gamma_l=0.1, limit_sum=False):
        """Calculate resonant Raman matrix

        Parameters
        ----------
        w_in: float
            Laser frequency in eV
        d_i: int
            Incoming polarization
        d_o: int
            Outgoing polarization
        gamma_l: float
            Line broadening in eV, (0.1eV=806rcm)
        limit_sum: bool
            Limit sum to occupied valence/unoccupied conduction states
            Use for debugging and testing. Don't use in production
        """

        raman_l = np.zeros((self.w_ph.shape[0]), dtype=complex)

        print(f"Calculating Raman spectrum: Laser frequency = {w_in} eV")

        # Loop over kpoints - this is parallelised
        for kpt in self.calc.wfs.kpt_u:
            # print(f"Rank {self.kd.comm.rank}: s={kpt.s}, k={kpt.k}")

            f_n = kpt.f_n / kpt.weight
            assert np.isclose(max(f_n), 1.0, atol=0.1)

            vs = np.arange(0, len(f_n))
            cs = np.arange(0, len(f_n))
            nc = 0
            nv = len(f_n)
            if limit_sum:  # good to test that we got arrays right
                vs = np.where(f_n >= 0.1)[0]
                cs = np.where(f_n < 0.9)[0]
                nv = max(vs) + 1  # VBM+1 index
                nc = min(cs)  # CBM index

            # Precalculate f * (1-f) term
            f_vc = np.outer(f_n[vs], 1.0 - f_n[cs])

            # Precalculate E-E term
            E_vc = np.zeros((len(vs), len(cs)), dtype=complex) + 1j * gamma_l
            for n in range(len(vs)):
                E_vc[n] += (kpt.eps_n[cs] - kpt.eps_n[n]) * Hartree

            # Obtain appropriate part of mom and g arrays
            pols = [d_i, d_o]
            mom_dnn = np.ascontiguousarray(self.mom_skvnm[kpt.s, kpt.k, pols])
            assert mom_dnn.shape[0] == 2
            g_lnn = np.ascontiguousarray(self.g_sqklnn[kpt.s, 0, kpt.k])

            # Raman contribution of this k-point
            raman_l += self.resonant_term(
                f_vc, E_vc, mom_dnn, g_lnn, nc, nv, w_in, self.w_ph
            )
        # Collect parallel contributions
        self.kd.comm.sum(raman_l)

        if self.kd.comm.rank == 0:
            print(f"Raman intensities per mode in {'xyz'[d_i]}{'xyz'[d_o]}")
            print("--------------------------")
            ff = "  Phonon {} with energy = {:4.2f} rcm: {:.4f}"
            for l in range(self.w_ph.shape[0]):
                print(ff.format(l, self.w_ph[l] / invcm, raman_l[l]))

        return raman_l

    def calculate_raman_tensor(self, w_in, gamma_l=0.1):
        """Calculate whole Raman tensor

        Parameters
        ----------
        w_in: float
            Laser frequency in eV
        gamma_l: float
            Line broadening in eV, (0.1eV=806rcm)
        """
        # If exist already, don't recompute
        for i in range(3):
            for j in range(3):
                with self.raman_cache.lock(f"{'xyz'[i]}{'xyz'[j]}") as handle:
                    if handle is None:
                        continue
                    R_l = self.calculate(w_in, i, j, gamma_l)
                    self.kd.comm.barrier()
                    handle.save(R_l)

    def nm_to_eV(self, laser_wave_length):
        return 1.239841e3 / laser_wave_length



import pytest
from gpaw.mpi import world
import time
import numpy as np

from gpaw.poisson import PoissonSolver
from gpaw.poisson_extravacuum import ExtraVacuumPoissonSolver
from gpaw.grid_descriptor import GridDescriptor


pytestmark = pytest.mark.skipif(world.size > 4,
                                reason='world.size > 4')


def test_poisson_poisson_extravacuum():
    do_output = False
    do_plot = False
    poissoneps = 1e-16

    if do_output:
        def equal(x, y, tol=0):
            res = {True: 'ok', False: 'not ok'}[abs(x - y) < tol]
            print('%.10e vs %.10e at %.10e is %s' % (x, y, tol, res))
    else:
        from gpaw.test import equal

    # Model grid
    N = 16
    N_c = np.array((1, 1, 3)) * N
    cell_c = N_c / float(N)
    gd = GridDescriptor(N_c, cell_c, False)

    # Construct model density
    coord_vg = gd.get_grid_point_coordinates()
    z_g = coord_vg[2, :]
    rho_g = gd.zeros()
    for z0 in [1, 2]:
        rho_g += 10 * (z_g - z0) * \
            np.exp(-20 * np.sum((coord_vg.T - np.array([.5, .5, z0])).T**2,
                                axis=0))

    if do_plot:
        big_rho_g = gd.collect(rho_g)
        if gd.comm.rank == 0:
            import matplotlib.pyplot as plt
            fig, ax_ij = plt.subplots(3, 4, figsize=(20, 10))
            ax_i = ax_ij.ravel()
            ploti = 0
            Ng_c = gd.get_size_of_global_array()
            plt.sca(ax_i[ploti])
            ploti += 1
            plt.pcolormesh(big_rho_g[Ng_c[0] / 2])
            plt.sca(ax_i[ploti])
            ploti += 1
            plt.plot(big_rho_g[Ng_c[0] / 2, Ng_c[1] / 2])

    def plot_phi(phi_g):
        if do_plot:
            big_phi_g = gd.collect(phi_g)
            if gd.comm.rank == 0:
                global ploti
                if ploti == 4:
                    ploti -= 2
                plt.sca(ax_i[ploti])
                ploti += 1
                plt.pcolormesh(big_phi_g[Ng_c[0] / 2])
                plt.sca(ax_i[ploti])
                ploti += 1
                plt.plot(big_phi_g[Ng_c[0] / 2, Ng_c[1] / 2])
                plt.ylim(np.array([-1, 1]) * 0.15)

    def poisson_solve(gd, rho_g, poisson):
        phi_g = gd.zeros()
        rho_g = rho_g
        t0 = time.time()
        npoisson = poisson.solve(phi_g, rho_g)
        t1 = time.time()
        if do_output:
            if gd.comm.rank == 0:
                print('Iterations: %s, Time: %.3f s' %
                      (str(npoisson), t1 - t0))
        return phi_g, npoisson

    def poisson_init_solve(gd, rho_g, poisson):
        poisson.set_grid_descriptor(gd)
        phi_g, npoisson = poisson_solve(gd, rho_g, poisson)
        plot_phi(phi_g)
        return phi_g, npoisson

    def compare(phi1_g, phi2_g, val, eps=np.sqrt(poissoneps)):
        big_phi1_g = gd.collect(phi1_g)
        big_phi2_g = gd.collect(phi2_g)
        if gd.comm.rank == 0:
            diff = np.max(np.absolute(big_phi1_g - big_phi2_g))
        else:
            diff = 0.0
        diff = np.array([diff])
        gd.comm.broadcast(diff, 0)
        equal(diff[0], val, eps)

    # Get reference from default poissonsolver
    poisson = PoissonSolver('fd', eps=poissoneps)
    phiref_g, npoisson = poisson_init_solve(gd, rho_g, poisson)

    # Test agreement with default
    poisson = ExtraVacuumPoissonSolver(
        N_c, PoissonSolver('fd', eps=poissoneps))
    phi_g, npoisson = poisson_init_solve(gd, rho_g, poisson)
    compare(phi_g, phiref_g, 0.0, 1e-15)

    # New reference with extra vacuum
    gpts = N_c * 4
    poisson = ExtraVacuumPoissonSolver(
        gpts, PoissonSolver('fd', eps=poissoneps))
    phi_g, npoisson = poisson_init_solve(gd, rho_g, poisson)
    # print poisson.get_description()
    compare(phi_g, phiref_g, 2.6485385144e-02)
    phiref_g = phi_g

    # Test with single coarsening
    poisson = ExtraVacuumPoissonSolver(
        gpts,
        PoissonSolver('fd', eps=poissoneps),
        PoissonSolver('fd', eps=poissoneps),
        1)
    phi_g, npoisson = poisson_init_solve(gd, rho_g, poisson)
    compare(phi_g, phiref_g, 1.5043946611e-04)

    # Test with two coarsenings
    poisson = ExtraVacuumPoissonSolver(
        gpts,
        PoissonSolver('fd', eps=poissoneps),
        PoissonSolver('fd', eps=poissoneps),
        2)
    phi_g, npoisson = poisson_init_solve(gd, rho_g, poisson)
    compare(phi_g, phiref_g, 1.2980906205e-03)

    # Test with cascaded single coarsenings
    poisson1 = ExtraVacuumPoissonSolver(gpts / 2,
                                        PoissonSolver('fd', eps=poissoneps),
                                        PoissonSolver('fd', eps=poissoneps), 1)
    poisson = ExtraVacuumPoissonSolver(gpts / 2, poisson1,
                                       PoissonSolver('fd', eps=poissoneps), 1)
    phi_g, npoisson = poisson_init_solve(gd, rho_g, poisson)
    # print poisson.get_description()
    compare(phi_g, phiref_g, 1.7086531461e-04)

    # Test auxgrid
    gpts = N_c * 4
    for coarses in [1, 2, 3]:
        for nn_refine in [1, 3]:
            for nn_laplace in [1, 3]:
                EVPS = ExtraVacuumPoissonSolver
                poisson = EVPS(gpts, PoissonSolver('fd', eps=poissoneps),
                               PoissonSolver('fd', eps=poissoneps), coarses,
                               use_aux_grid=False,
                               nn_refine=nn_refine, nn_laplace=nn_laplace)
                phiref_g, npoisson = poisson_init_solve(gd, rho_g, poisson)

                poisson = EVPS(gpts, PoissonSolver('fd', eps=poissoneps),
                               PoissonSolver('fd', eps=poissoneps), coarses,
                               use_aux_grid=True,
                               nn_refine=nn_refine, nn_laplace=nn_laplace)
                phi_g, npoisson = poisson_init_solve(gd, rho_g, poisson)
                compare(phi_g, phiref_g, 0.0, 1e-14)

    if do_plot:
        if gd.comm.rank == 0:
            plt.show()



import numpy as np
import pytest
from gpaw import GPAW
from gpaw.ibz2bz import IBZ2BZMaps, get_overlap, get_overlap_coefficients
import gpaw.mpi as mpi
from gpaw.test.conftest import response_band_cutoff


@pytest.mark.later
@pytest.mark.serial
@pytest.mark.parametrize('gs', ['fancy_si_pw',
                                'al_pw',
                                'fe_pw',
                                'co_pw',
                                'gaas_pw',
                                'v2br4_pw',
                                'srvo3_pw'])
def test_ibz2bz(in_tmp_dir, gpw_files, gs):
    """ Tests gpaw.ibz2bz.py
    Tests functionalities to take wavefunction and projections from
    ibz to full bz by comparing calculations with and without symmetry.
    """

    atol = 5e-03  # Tolerance when comparing wfs and projections
    atol_eig = 1e-04  # Tolerance when comparing eigenvalues
    atol_deg = 5e-3  # Tolerance for checking degenerate states

    # Loading calc with symmetry
    calc = GPAW(gpw_files[gs],
                communicator=mpi.serial_comm)
    wfs = calc.wfs
    nconv = response_band_cutoff[gs]

    # setting basic stuff
    nbands = wfs.bd.nbands if nconv == -1 else nconv
    nbzk = wfs.kd.nbzkpts
    ibz2bz = IBZ2BZMaps.from_calculator(calc)

    # Loading calc without symmetry
    calc_nosym = GPAW(gpw_files[gs + '_nosym'],
                      communicator=mpi.serial_comm)
    wfs_nosym = calc_nosym.wfs

    # Check some basic stuff
    assert wfs_nosym.kd.nbzkpts == wfs_nosym.kd.nibzkpts

    # Loop over spins and k-points
    for s in range(wfs.nspins):
        for K in range(nbzk):
            ik = wfs.kd.bz2ibz_k[K]  # IBZ k-point

            # Check so that BZ kpoints are the same
            assert np.allclose(wfs.kd.bzk_kc[K], wfs_nosym.kd.bzk_kc[K])
            assert np.allclose(wfs_nosym.kd.ibzk_kc[K], wfs_nosym.kd.bzk_kc[K])

            # Get data for calc without symmetry at BZ kpt K
            eps_n_nosym, ut_nR_nosym, proj_nosym, dO_aii_nosym = \
                get_ibz_data_from_wfs(wfs_nosym, nbands, K, s)

            # Get data for calc with symmetry at ibz kpt ik
            eps_n, ut_nR, proj, dO_aii = get_ibz_data_from_wfs(wfs,
                                                               nbands,
                                                               ik, s)

            # Map projections and u:s from ik to K
            proj_sym = ibz2bz[K].map_projections(proj)
            ut_nR_sym = np.array([ibz2bz[K].map_pseudo_wave_to_BZ(
                ut_nR[n]) for n in range(nbands)])

            # Check so that eigenvalues are the same
            assert np.allclose(eps_n[:nbands],
                               eps_n_nosym[:nbands],
                               atol=atol_eig)

            # Check so that overlaps are the same for both calculations
            assert equal_dicts(dO_aii,
                               dO_aii_nosym,
                               atol)

            # Here starts the actual test
            # Loop over all bands
            n = 0
            while n < nbands:
                dim = find_degenerate_subspace(eps_n, n, nbands, atol_deg)
                if dim == 1:

                    # First check untransformed quantities for ibz k-points
                    if np.allclose(wfs.kd.bzk_kc[K],
                                   wfs.kd.ibzk_kc[ik]):
                        # Compare untransformed projections
                        compare_projections(proj, proj_nosym, n, atol)
                        # Compare untransformed wf:s
                        assert np.allclose(abs(ut_nR[n]),
                                           abs(ut_nR_nosym[n]), atol=atol)

                    # Then check so that absolute value of transformed
                    # projections are the same
                    compare_projections(proj_sym, proj_nosym, n, atol)

                    # Check so that periodic part of pseudo is same,
                    # up to a phase
                    assert np.allclose(abs(ut_nR_sym[n]),
                                       abs(ut_nR_nosym[n]), atol=atol)

                # For degenerate states check transformation
                # matrix is unitary,
                # For non-degenerate states check so that all-electron wf:s
                # are the same up to phase
                bands = range(n, n + dim)

                check_all_electron_wfs(bands, wfs.gd,
                                       ut_nR_sym, ut_nR_nosym,
                                       proj_sym, proj_nosym, dO_aii,
                                       atol)
                n += dim


def equal_dicts(dict_1, dict_2, atol):
    """ Checks so that two dicts with np.arrays are
    equal"""
    assert len(dict_1.keys()) == len(dict_2.keys())
    for key in dict_1:
        # Make sure the dictionaries contain the same set of keys
        if key not in dict_2:
            return False
        # Make sure that the arrays are identical
        if not np.allclose(dict_1[key], dict_2[key], atol=atol):
            return False
    return True


def find_degenerate_subspace(eps_n, n_start, nbands, atol_eig):
    # Find degenerate eigenvalues
    n = n_start
    dim = 1
    while n < nbands - 1 and abs(eps_n[n] - eps_n[n + 1]) < atol_eig:
        dim += 1
        n += 1
    return dim


def compare_projections(proj_sym, proj_nosym, n, atol):
    # compares so that projections at given k and band index n
    # differ by at most a phase
    for a, P_ni in proj_sym.items():
        for j in range(P_ni.shape[1]):
            # Check so that absolute values of projections are the same
            assert np.isclose(abs(P_ni[n, j]),
                              abs(proj_nosym[a][n, j]),
                              atol=atol)


def check_all_electron_wfs(bands, gd, ut1_nR, ut2_nR,
                           proj_sym, proj_nosym, dO_aii,
                           atol):
    """sets up transformation matrix between symmetry
       transformed u:s and normal u:s in degenerate subspace
       and asserts that it is unitary. It also checks that
       the pseudo wf:s transform according to the same
       transformation.

       Let |ψ^1_i> denote the all electron wavefunctions
       from the calculation with symmetry and |ψ^2_i>
       the corresponding wavefunctions from the calculation
       without symmetry.
       If the set {|ψ^1_i>} span the same subspace as the set
       {|ψ^2_i>} they fulfill the following where summation
       over repeated indexes is assumed:

       |ψ^2_i> = |ψ^1_k> <ψ^1_k |ψ^2_i> == M_ki |ψ^1_k>
       and M_ki = <ψ^1_k |ψ^2_i>  is a unitary transformation.
       M_ki is only unitary if the two sets of wfs span the
       same subspace.

    Parameters
    ---------
    bands: list of ints
         band indexes in degenerate subspace
    ut1_nR: np.array
    ut2_nR: np.array
        Periodic part of pseudo wave function for two calculations
    proj_sym: Projections object
    proj_nosym: Projections object
        Projections for two calculations
    dO_aii: dict with np.arrays
       see ibz2bz.get_overlap_coefficients
    dv:     float
            calc.wfs.gd.dv
    atol: float
       absolute tolerance when comparing arrays
    """
    M_nn = get_overlap(bands,
                       gd,
                       ut1_nR,
                       ut2_nR,
                       proj_sym,
                       proj_nosym,
                       dO_aii)

    # Check so that transformation matrix is unitary
    MMdag_nn = M_nn @ M_nn.T.conj()
    assert np.allclose(np.eye(len(MMdag_nn)), MMdag_nn, atol=atol)

    # Check so that M_nn transforms pseudo wf:s, see docs
    ut2_from_transform_nR = np.einsum('ji,jklm->iklm', M_nn, ut1_nR[bands])
    assert np.allclose(ut2_from_transform_nR, ut2_nR[bands], atol=atol)


def get_ibz_data_from_wfs(wfs, nbands, ik, s):
    """ gets data at ibz k-point ik
    """
    # get energies and wfs
    kpt = wfs.kpt_qs[ik][s]
    psit_nG = kpt.psit_nG
    eps_n = kpt.eps_n

    # Get periodic part of pseudo wfs
    ut_nR = np.array([wfs.pd.ifft(
        psit_nG[n], ik) for n in range(nbands)])

    # Get projections
    proj = kpt.projections.new(nbands=nbands, bcomm=None)
    proj.array[:] = kpt.projections.array[:nbands]

    # Get overlap coefficients from PAW setups
    dO_aii = get_overlap_coefficients(wfs)
    return eps_n, ut_nR, proj, dO_aii



from collections import deque
from inspect import signature

import numpy as np
from ase.calculators.calculator import InputError
from ase.units import Bohr, Ha

from gpaw.mpi import broadcast_float


def get_criterion(name):
    """Returns one of the pre-specified criteria by it's .name attribute,
    and raises sensible error if missing."""
    # All built-in criteria should be in this list.
    criteria = [Energy, Density, Eigenstates, Forces, WorkFunction,
                MinIter, MaxIter]
    criteria = {c.name: c for c in criteria}
    try:
        return criteria[name]
    except KeyError:
        known = ', '.join(f'{key!r}' for key in criteria)
        msg = (
            f'The convergence keyword "{name}" was supplied, which we do not '
            'know how to handle. If this is a typo, please correct '
            f'(known keywords are {known}). If this'
            ' is a user-written convergence criterion, it cannot be '
            'imported with this function; please see the GPAW manual for '
            'details.')
        raise InputError(msg)


def dict2criterion(dictionary):
    """Converts a dictionary to a convergence criterion.

    The dictionary can either be that generated from 'todict'; that is like
    {'name': 'energy', 'tol': 0.005, 'n_old': 3}. Or from user-specified
    shortcut like {'energy': 0.005} or {'energy': (0.005, 3)}, or a
    combination like {'energy': {'name': 'energy', 'tol': 0.005, 'n_old': 3}.
    """
    d = dictionary.copy()
    if 'name' in d:  # from 'todict'
        name = d.pop('name')
        ThisCriterion = get_criterion(name)
        return ThisCriterion(**d)
    assert len(d) == 1
    name = list(d.keys())[0]
    if isinstance(d[name], dict) and 'name' in d[name]:
        return dict2criterion(d[name])
    ThisCriterion = get_criterion(name)
    return ThisCriterion(*[d[name]])


def check_convergence(criteria, ctx):
    entries = {}  # for log file, per criteria
    converged_items = {}  # True/False, per criteria
    override_others = False
    converged = True
    for name, criterion in criteria.items():
        if not criterion.calc_last:
            ok, entry = criterion(ctx)
            if criterion.override_others:
                if ok:
                    override_others = True
            else:
                converged = converged and ok
            converged_items[name] = ok
            entries[name] = entry

    for name, criterion in criteria.items():
        if criterion.calc_last:
            if converged:
                ok, entry = criterion(ctx)
                converged &= ok
                converged_items[name] = ok
                entries[name] = entry
            else:
                converged_items[name] = False
                entries[name] = ''

    # Converged?
    return converged or override_others, converged_items, entries


class Criterion:
    """Base class for convergence criteria.

    Automates the creation of the __repr__ and todict methods for generic
    classes. This will work for classes that save all arguments directly,
    like __init__(self, a, b):  --> self.a = a, self.b = b. The todict
    method requires the class have a self.name attribute. All criteria
    (subclasses of Criterion) must define self.name, self.tablename,
    self.description, self.__init__, and self.__call___. See the online
    documentation for details.
    """
    # If calc_last is True, will only be checked after all other (non-last)
    # criteria have been met.
    calc_last = False
    override_others = False
    description: str

    def __repr__(self):
        parameters = signature(self.__class__).parameters
        s = ', '.join([str(getattr(self, p)) for p in parameters])
        return self.__class__.__name__ + '(' + s + ')'

    def todict(self):
        d = {'name': self.name}
        parameters = signature(self.__class__).parameters
        for parameter in parameters:
            d[parameter] = getattr(self, parameter)
        return d

    def reset(self):
        pass


# Built-in criteria follow. Make sure that any new criteria added below
# are also added to to the list in get_criterion() so that it can import
# them correctly by name.


class Energy(Criterion):
    """A convergence criterion for the total energy.

    Parameters:

    tol: float
        Tolerance for conversion; that is the maximum variation among the
        last n_old values of the (extrapolated) total energy.
    n_old: int
        Number of energy values to compare. I.e., if n_old is 3, then this
        compares the peak-to-peak difference among the current total energy
        and the two previous.
    relative: bool
        Use total energy [eV] or total energy relative to number of
        valence electrons [eV/(valence electron)].
    """
    name = 'energy'
    tablename = 'energy'

    def __init__(self, tol: float, *, n_old: int = 3, relative: bool = True):
        self.tol = tol
        self.n_old = n_old
        self.relative = relative
        self.description = (
            f'Maximum [total energy] change in last {self.n_old} cyles: '
            f'{self.tol:g} eV')
        if relative:
            self.description += ' / valence electron'

    def reset(self):
        self._old = deque(maxlen=self.n_old)

    def __call__(self, context):
        """Should return (bool, entry), where bool is True if converged and
        False if not, and entry is a <=5 character string to be printed in
        the user log file."""
        # Note the previous code was calculating the peak-to-
        # peak energy difference on e_total_free, while reporting
        # e_total_extrapolated in the SCF table (logfile). I changed it to
        # use e_total_extrapolated for both. (Should be a miniscule
        # difference, but more consistent.)
        total_energy = context.ham.e_total_extrapolated * Ha
        if context.wfs.nvalence == 0 or not self.relative:
            energy = total_energy
        else:
            energy = total_energy / context.wfs.nvalence
        self._old.append(energy)  # Pops off >3!
        error = np.inf
        if len(self._old) == self._old.maxlen:
            error = np.ptp(self._old)
        converged = error < self.tol
        entry = ''
        if np.isfinite(energy):
            entry = '{:11.6f}'.format(total_energy)
        return converged, entry


class Density(Criterion):
    """A convergence criterion for the electron density.

    Parameters:

    tol : float
        Tolerance for conversion; that is the maximum change in the electron
        density, calculated as the integrated absolute value of the density
        change, normalized per valence electron. [electrons/(valence electron)]
    """
    name = 'density'
    tablename = 'dens'

    def __init__(self, tol):
        self.tol = tol
        self.description = ('Maximum integral of absolute [dens]ity change: '
                            '{:g} electrons / valence electron'
                            .format(self.tol))

    def __call__(self, context):
        """Should return (bool, entry), where bool is True if converged and
        False if not, and entry is a <=5 character string to be printed in
        the user log file."""
        if context.dens.fixed:
            return True, ''
        nv = context.wfs.nvalence
        if nv == 0:
            return True, ''
        # Make sure all agree on the density error.
        error = broadcast_float(context.dens.error, context.wfs.world) / nv
        converged = (error < self.tol)
        if (error is None or np.isinf(error) or error == 0):
            entry = ''
        else:
            entry = '{:+5.2f}'.format(np.log10(error))
        return converged, entry


class Eigenstates(Criterion):
    """A convergence criterion for the eigenstates.

    Parameters:

    tol : float
        Tolerance for conversion; that is the maximum change in the
        eigenstates, calculated as the integration of the square of the
        residuals of the Kohn--Sham equations, normalized per valence
        electron. [eV^2/(valence electron)]
    """
    name = 'eigenstates'
    tablename = 'eigst'

    def __init__(self, tol):
        self.tol = tol
        self.description = ('Maximum integral of absolute [eigenst]ate '
                            'change: {:g} eV^2 / valence electron'
                            .format(self.tol))

    def __call__(self, context):
        """Should return (bool, entry), where bool is True if converged and
        False if not, and entry is a <=5 character string to be printed in
        the user log file."""
        if context.wfs.nvalence == 0:
            return True, ''
        error = self.get_error(context)
        converged = (error < self.tol)
        if (context.wfs.nvalence == 0 or error == 0 or np.isinf(error)):
            entry = ''
        else:
            entry = '{:+5.2f}'.format(np.log10(error))
        return converged, entry

    def get_error(self, context):
        """Returns the raw error."""
        return context.wfs.eigensolver.error * Ha**2 / context.wfs.nvalence


class Forces(Criterion):
    """A convergence criterion for the forces.

    Parameters:

    atol : float
        Absolute tolerance for convergence; that is, the force on each atom
        is compared with its force from the previous iteration, and the change
        in each atom's force is calculated as an l2-norm
        (Euclidean distance). The atom with the largest norm must be less
        than tol. [eV/Angstrom]
    rtol : float
        Relative tolerance for convergence. The difference in the l2-norm of
        force on each atom is calculated, and convergence is achieved when
        the largest difference between two iterations is rtol * max force.
    calc_last : bool
        If True, calculates forces last; that is, it waits until all other
        convergence criteria are satisfied before checking to see if the
        forces have converged. (This is more computationally efficient.)
        If False, checks forces at each SCF step.
    """
    name = 'forces'
    tablename = 'force'

    def __init__(self, atol, rtol=np.inf, calc_last=True):
        self.atol = atol
        self.rtol = rtol
        self.description = ('Maximum change in the atomic [forces] across '
                            f'last 2 cycles: {self.atol} eV/Ang OR\n'
                            'Maximum error relative to the maximum '
                            f'force is below {self.rtol}')
        self.calc_last = calc_last
        self.reset()

    def __call__(self, context):
        """Should return (bool, entry), where bool is True if converged and
        False if not, and entry is a <=5 character string to be printed in
        the user log file."""

        # criterion is off; backwards compatibility
        if np.isinf(self.atol) and np.isinf(self.rtol):
            return True, ''
        F_av = context.calculate_forces()
        F_av *= Ha / Bohr
        error = np.inf
        max_force = np.max(np.linalg.norm(F_av, axis=1))
        if self.old_F_av is not None:
            error = ((F_av - self.old_F_av)**2).sum(1).max()**0.5
        self.old_F_av = F_av

        if np.isfinite(self.rtol):
            error_threshold = min(self.atol, self.rtol * max_force)
        else:
            # Avoid possible inf * 0.0:
            error_threshold = self.atol
        converged = error < error_threshold

        entry = ''
        if np.isfinite(error):
            if error:
                entry = '{:+5.2f}'.format(np.log10(error))
            else:
                entry = '-inf'
        return converged, entry

    def reset(self):
        self.old_F_av = None


class WorkFunction(Criterion):
    """A convergence criterion for the work function.

    Parameters:

    tol : float
        Tolerance for conversion; that is the maximum variation among the
        last n_old values of either work function. [eV]
    n_old : int
        Number of work functions to compare. I.e., if n_old is 3, then this
        compares the peak-to-peak difference among the current work
        function and the two previous.
    """
    name = 'work function'
    tablename = 'wkfxn'

    def __init__(self, tol=0.005, n_old=3):
        self.tol = tol
        self.n_old = n_old
        self.description = ('Maximum change in the last {:d} '
                            'work functions [wkfxn]: {:g} eV'
                            .format(n_old, tol))

    def reset(self):
        self._old = deque(maxlen=self.n_old)

    def __call__(self, context):
        """Should return (bool, entry), where bool is True if converged and
        False if not, and entry is a <=5 character string to be printed in
        the user log file."""
        workfunctions = context.ham.get_workfunctions(context.wfs)
        workfunctions = Ha * np.array(workfunctions)
        self._old.append(workfunctions)  # Pops off >3!
        if len(self._old) == self._old.maxlen:
            error = max(np.ptp(self._old, axis=0))
        else:
            error = np.inf
        converged = (error < self.tol)
        if error < np.inf:
            entry = '{:+5.2f}'.format(np.log10(error))
        else:
            entry = ''
        return converged, entry


class MinIter(Criterion):
    """A convergence criterion that enforces a minimum number of iterations.

    Parameters:

    n : int
        Minimum number of iterations that must be complete before
        the SCF cycle exits.
    """
    calc_last = False
    name = 'minimum iterations'
    tablename = 'minit'

    def __init__(self, n):
        self.n = n
        self.description = f'Minimum number of iterations [minit]: {n}'

    def __call__(self, context):
        converged = context.niter >= self.n
        entry = '{:d}'.format(context.niter)
        return converged, entry


class MaxIter(Criterion):
    """A convergence criterion that enforces a maximum number of iterations.

    Parameters:

    n : int
        Maximum number of iterations that must be complete before
        the SCF cycle exits.
    """
    calc_last = False
    name = 'maximum iterations'
    tablename = 'maxit'
    override_others = True

    def __init__(self, n):
        self.n = n
        self.description = f'Maximum number of iterations [minit]: {n}'

    def __call__(self, context):
        converged = context.niter >= self.n
        entry = '{:d}'.format(context.niter)
        return converged, entry



"""This calculation has the following structure.

1) Calculate the ground state of diamond.
2) Calculate the band structure of diamond in order to obtain
   accurate KS band gap.
3) Calculate the potential discontinuity using the ground state calculator
   and the accurate band gap.
4) Apply the discontinuity to CBM using the band structure calculator.

Compare to reference.
"""
import pytest
from ase.build import bulk
from gpaw import GPAW, Davidson, Mixer
from gpaw.mpi import world
from gpaw.calculator import DeprecatedParameterWarning


@pytest.mark.gllb
@pytest.mark.libxc
@pytest.mark.parametrize('deprecated_syntax', [False, True])
def test_gllb_diamond(in_tmp_dir, deprecated_syntax):
    xc = 'GLLBSC'
    KS_gap_ref = 4.180237125868162
    QP_gap_ref = 5.469387490357182
    KS_gap_direct_ref = 5.682571428480924
    QP_gap_direct_ref = 7.452331561454596
    dxc_ave_ref = 1.33769712051036
    # M. Kuisma et. al, https://doi.org/10.1103/PhysRevB.82.115106
    #     C: KS gap 4.14 eV, QP gap 5.41eV, expt. 5.48 eV

    # Calculate ground state
    atoms = bulk('C', 'diamond', a=3.567)
    # We want sufficiently many grid points that the calculator
    # can use wfs.world for the finegd, to test that part of the code.
    calc = GPAW(mode='fd',
                h=0.2,
                kpts=(4, 4, 4),
                xc=xc,
                nbands=8,
                mixer=Mixer(0.5, 5, 10.0),
                parallel=dict(domain=min(world.size, 2), band=1),
                eigensolver=Davidson(niter=2),
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()

    if deprecated_syntax:
        with pytest.warns(DeprecationWarning):
            # Calculate the discontinuity directly
            response = calc.hamiltonian.xc.xcs['RESPONSE']
            response.calculate_delta_xc()
            KS_gap, dxc = response.calculate_delta_xc_perturbation()
    else:
        # Calculate the discontinuity directly
        homo, lumo = calc.get_homo_lumo()
        response = calc.hamiltonian.xc.response
        dxc_pot = response.calculate_discontinuity_potential(homo, lumo)
        KS_gap, dxc = response.calculate_discontinuity(dxc_pot)

    QP_gap = KS_gap + dxc
    print('KS gap direct', KS_gap)
    print('QP gap direct', QP_gap)
    assert KS_gap == pytest.approx(KS_gap_direct_ref, abs=1e-4)
    assert QP_gap == pytest.approx(QP_gap_direct_ref, abs=1e-4)

    if deprecated_syntax:
        with pytest.warns(DeprecatedParameterWarning):
            from gpaw import restart

            calc.write('gs.gpw')
            # Calculate accurate KS-band gap from band structure
            atoms, bs_calc = restart('gs.gpw',
                                     fixdensity=True,
                                     kpts={'path': 'GX', 'npoints': 12},
                                     symmetry='off',
                                     nbands=8,
                                     convergence={'bands': 6},
                                     eigensolver=Davidson(niter=4),
                                     txt='bs.out')
            atoms.get_potential_energy()

        # Get the accurate KS-band gap
        homolumo = bs_calc.get_homo_lumo()
        homo, lumo = homolumo
    else:
        # Calculate accurate KS-band gap from band structure
        bs_calc = calc.fixed_density(kpts={'path': 'GX', 'npoints': 12},
                                     symmetry='off',
                                     nbands=8,
                                     convergence={'bands': 6},
                                     eigensolver=Davidson(niter=4),
                                     txt='bs.out')

        # Get the accurate KS-band gap
        homo, lumo = bs_calc.get_homo_lumo()

    if deprecated_syntax:
        with pytest.warns(DeprecationWarning):
            from ase.units import Ha
            from gpaw.xc.gllb.c_response import ResponsePotential

            # Calculate the discontinuity potential with accurate band gap
            response = calc.hamiltonian.xc.xcs['RESPONSE']
            response.calculate_delta_xc(homolumo=homolumo / Ha)

            # Calculate the discontinuity using the band structure calculator
            bs_response = bs_calc.hamiltonian.xc.xcs['RESPONSE']
            # Copy response potential to avoid recalculation
            pot = ResponsePotential(response,
                                    response.Dxc_vt_sG,
                                    None,
                                    response.Dxc_D_asp,
                                    response.Dxc_Dresp_asp)
            pot.redistribute(bs_response)
            bs_response.Dxc_vt_sG = pot.vt_sG
            bs_response.Dxc_D_asp = pot.D_asp
            bs_response.Dxc_Dresp_asp = pot.Dresp_asp
            KS_gap, dxc = bs_response.calculate_delta_xc_perturbation()
    else:
        # Calculate the discontinuity potential with accurate band gap
        response = calc.hamiltonian.xc.response
        dxc_pot = response.calculate_discontinuity_potential(homo, lumo)

        # Calculate the discontinuity using the band structure calculator
        bs_response = bs_calc.hamiltonian.xc.response
        KS_gap, dxc = bs_response.calculate_discontinuity(dxc_pot)

        # Test alternative (testing) implementation
        if world.size == 1:
            with pytest.warns(Warning):
                dxc_ave = bs_response.calculate_discontinuity_from_average(
                    dxc_pot, 0, True)
            assert dxc_ave == pytest.approx(dxc_ave_ref, abs=1e-4)

    QP_gap = KS_gap + dxc
    print('KS gap', KS_gap)
    print('QP gap', QP_gap)
    assert KS_gap == pytest.approx(lumo - homo, abs=1e-10)
    assert KS_gap == pytest.approx(KS_gap_ref, abs=1e-4)
    assert QP_gap == pytest.approx(QP_gap_ref, abs=1e-4)



from __future__ import annotations
from gpaw.core.arrays import DistributedArrays as XArray
from gpaw.core import UGArray


class Hamiltonian:
    def apply(self,
              vt_sR: UGArray,
              dedtaut_sR: UGArray | None,
              psit_nG: XArray,
              out: XArray,
              spin: int) -> XArray:
        self.apply_local_potential(vt_sR[spin], psit_nG, out)
        if dedtaut_sR is not None:
            self.apply_mgga(dedtaut_sR[spin], psit_nG, out)
        return out

    def apply_local_potential(self,
                              vt_R: UGArray,
                              psit_nG: XArray,
                              out: XArray
                              ) -> None:
        raise NotImplementedError

    def apply_mgga(self,
                   dedtaut_R: UGArray,
                   psit_nG: XArray,
                   vt_nG: XArray) -> None:
        raise NotImplementedError

    def create_preconditioner(self, blocksize):
        raise NotImplementedError



import numpy as np

import ase.io.ulm as ulm

from gpaw.io import Writer
from gpaw.lcaotddft.observer import TDDFTObserver


class LineDensityReader(object):
    def __init__(self, filename):
        self.reader = ulm.Reader(filename)
        tag = self.reader.get_tag()
        if tag != LineDensityWriter.ulmtag:
            raise RuntimeError('Unknown tag %s' % tag)
        self.filename = filename

    def __getattr__(self, attr):
        return getattr(self.reader, attr)

    @property
    def kick_strength(self):
        for ri in range(1, len(self.reader)):
            r = self.reader[ri]
            if r.action == 'kick':
                return r.kick_strength

    def read(self, skip_duplicates=True, zero_pad='both'):
        assert zero_pad in ['both', False, None]
        time_t = []
        rho_tx = []
        time0 = -np.inf
        for ri in range(1, len(self.reader)):
            r = self.reader[ri]
            time = r.time
            if time > time0 or not skip_duplicates:
                time_t.append(time)
                rho_x = r.rho_x
                if zero_pad == 'both':
                    rho_x = np.concatenate(([0], rho_x, [0]))
                rho_tx.append(rho_x)
                time0 = time
        time_t = np.array(time_t)
        rho_tx = np.array(rho_tx)
        return time_t, rho_tx

    def close(self):
        self.reader.close()


class LineDensityWriter(TDDFTObserver):
    version = 1
    ulmtag = 'LineDensity'

    def __init__(self, paw, filename, c=0, density_type='comp', interval=1):
        TDDFTObserver.__init__(self, paw, interval)
        if paw.niter == 0:
            self.density_type = density_type
            self.c = c
            density = paw.density
            if 'coarse' in density_type:
                gd = density.gd
            else:
                gd = density.finegd
            h_v = gd.h_cv[self.c]

            self.writer = Writer(filename, paw.world, mode='w',
                                 tag=self.__class__.ulmtag)
            self.writer.write(version=self.__class__.version,
                              density_type=self.density_type,
                              c=self.c, h_v=h_v)
            self.writer.sync()
        else:
            # Read settings
            reader = LineDensityReader(filename)
            self.density_type = reader.density_type
            self.c = reader.c
            h_v = reader.h_v

            # Append to earlier file
            self.writer = Writer(filename, paw.world, mode='a',
                                 tag=self.__class__.ulmtag)
        self.dx = np.linalg.norm(h_v)
        assert self.density_type in ['comp', 'pseudo', 'pseudocoarse'], \
            'Unknown density type: {}'.format(self.density_type)

    def _update(self, paw):
        # Write metadata to main writer
        self.writer.write(niter=paw.niter, time=paw.time, action=paw.action)
        if paw.action == 'kick':
            self.writer.write(kick_strength=paw.kick_strength)

        density = paw.density

        if self.density_type == 'comp':
            rho_g = density.rhot_g
            gd = density.finegd
        elif self.density_type == 'pseudo':
            rho_g = density.nt_sg.sum(axis=0)
            gd = density.finegd
        elif self.density_type == 'pseudocoarse':
            rho_g = density.nt_sG.sum(axis=0)
            gd = density.gd
        else:
            raise RuntimeError('Unknown density type: %s' % self.density_type)

        rho_g = gd.collect(rho_g)
        # TODO: instead of collecting on master,
        # do integration in each domain and gather/sum to master
        if gd.comm.rank == 0:
            sum_axis = tuple({0, 1, 2}.difference({self.c}))
            rho_x = rho_g.sum(axis=sum_axis) * (gd.dv / self.dx)
        else:
            rho_x = None
        self.writer.write(rho_x=rho_x)

        # Sync the main writer
        self.writer.sync()

    def __del__(self):
        self.writer.close()



import sys
from pathlib import Path

import pytest

from gpaw import GPAW, PW


@pytest.mark.later
@pytest.mark.ci
def test_new_calculator(in_tmp_dir):
    """Test the GPAW.new() method."""

    params = dict(
        mode=PW(200),
        xc='LDA',
        nbands=8,
        kpts={'size': (4, 4, 4), 'gamma': True})

    modification_m = [
        dict(mode='fd'),
        dict(xc='PBE'),
        dict(nbands=10),
        dict(kpts={'size': (4, 4, 4)}),
        dict(kpts={'size': (3, 3, 3)}, xc='PBE'),
    ]

    calc0 = GPAW(**params, txt='calc0.txt')

    for m, modification in enumerate(modification_m):
        if m == 0:
            # Don't give a new txt file
            calc = calc0.new(**modification)
            check_file_handles(calc0, calc)
        else:
            txt = f'calc{m}.txt'
            calc = calc0.new(**modification, txt=txt)
            check_file_handles(calc0, calc, txt=txt)

        check_calc(calc, params, modification, world=calc.world)


def check_file_handles(calc0, calc, txt=None):
    assert calc.log.world.rank == calc0.log.world.rank

    if calc.log.world.rank == 0:
        # We never want to reuse the output file
        assert calc.log._fd is not calc0.log._fd

        if txt is None:
            # When no txt is specified, the new calculator should log its
            # output in stdout
            assert calc.log._fd is sys.stdout
        else:
            # Check that the new calculator log file handle was updated
            # appropriately
            assert Path(calc.log._fd.name).name == txt


def check_calc(calc, params, modification, *, world):
    desired_params = params.copy()
    desired_params.update(modification)

    for param, value in desired_params.items():
        assert calc.parameters[param] == value

    # Check that the communicator is reused
    assert calc.world is world



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""This module defines a ``KPoint`` class."""
from typing import Optional
from gpaw.projections import Projections


class KPoint:
    """Class for a single k-point.

    The KPoint class takes care of all wave functions for a
    certain k-point and a certain spin.

    Attributes:

    eps_n: float ndarray
        Eigenvalues.
    f_n: float ndarray
        Occupation numbers.  The occupation numbers already include the
        k-point weight in supercell calculations.
    """

    def __init__(self,
                 weightk: float,
                 weight: float,
                 s: int,
                 k: int,
                 q: int,
                 phase_cd=None):
        """Construct k-point object.

        Parameters:

        weightk:
            Weight of this k-point.
        weight:
            Old confusing weight.
        s: int
            Spin index: up or down (0 or 1).
        k: int
            k-point IBZ-index.
        q: int
            local k-point index.
        phase_cd: complex ndarray
            Bloch phase-factors for translations - axis c=0,1,2
            and direction d=0,1.
        """

        self.weightk = weightk  # pure k-point weight
        self.weight = weight  # old confusing weight
        self.s = s  # spin index
        self.k = k  # k-point index
        self.q = q  # local k-point index
        self.phase_cd = phase_cd

        self.eps_n = None
        self.f_n = None
        self._projections: Optional[Projections] = None

        # Only one of these two will be used:
        self.psit = None  # UniformGridMatrix/PWExpansionMatrix
        self.C_nM = None  # LCAO coefficients for wave functions

        # LCAO stuff:
        self.rho_MM = None
        self.S_MM = None
        self.T_MM = None

    def __repr__(self):
        return (f'KPoint(weight={self.weight}, weightk={self.weightk}, '
                f's={self.s}, k={self.k}, q={self.q})')

    @property
    def projections(self) -> Projections:
        assert self._projections is not None
        return self._projections

    @projections.setter
    def projections(self, value):
        self._projections = value

    @property
    def P_ani(self):
        if self._projections is not None:
            return {a: P_ni for a, P_ni in self.projections.items()}

    @property
    def psit_nG(self):
        if self.psit is not None:
            return self.psit.array



import os

import sys

import urllib.request
import urllib.parse
import urllib.error

import zipfile

import csv

import numpy as np

from ase.test import NotAvailable

from ase import units

from ase.test.tasks.dcdft import DeltaCodesDFTTask as Task

dir = 'Delta'

if len(sys.argv) == 1:
    tag = None
    reffile = os.path.join(dir, 'WIEN2k.txt')
else:
    if len(sys.argv) == 3:
        tag = sys.argv[1]
        reffile = sys.argv[2]
    else:
        tag = sys.argv[1]
        reffile = os.path.join(dir, 'WIEN2k.txt')

src = 'https://molmod.ugent.be/sites/default/files/Delta_v3-0.zip'
name = os.path.basename(src)
if not os.path.exists(dir):
    os.makedirs(dir)
os.chdir(dir)
if not os.path.exists('calcDelta.py'):
    try:
        resp = urllib.request.urlopen(src)
        urllib.request.urlretrieve(src, filename=name)
        z = zipfile.ZipFile(name)
        try:  # new in 2.6
            z.extractall()
        except AttributeError:
            # http://stackoverflow.com/questions/7806563/how-to-unzip-a-zip-file-with-python-2-4
            for f in z.namelist():
                fd = open(f, "w")
                fd.write(z.read(f))
                fd.close()
        # AttributeError if unzip not found
    except (urllib.error.HTTPError, AttributeError):
        raise NotAvailable('Retrieval of zip failed')
os.chdir('..')

task = Task(
    tag=tag,
    use_lock_files=True)

# header
h = ['#element', 'V0', 'B0', 'B1']

if not os.path.exists('%s_raw.csv' % tag):
    # read calculated results from json file and write into csv
    task.read()
    task.analyse()

    f1 = open('%s_raw.csv' % tag, 'wb')
    csvwriter1 = csv.writer(f1)
    csvwriter1.writerow(h)

    for n in task.collection.names:
        row = [n]
        if n in task.data.keys():
            try:
                v = task.data[n]['dcdft volume']
                b0 = task.data[n]['dcdft B0'] / (units.kJ * 1e-24)
                b1 = task.data[n]['dcdft B1']
                row.extend([v, b0, b1])
            except KeyError:  # completely failed to find eos minimum
                row.extend(['N/A', 'N/A', 'N/A'])
        else:
            # element not calculated
            row.extend(['N/A', 'N/A', 'N/A'])
        if 'N/A' not in row:
            csvwriter1.writerow(row)
    f1.close()

# read raw results
csvreader1 = csv.reader(open('%s_raw.csv' % tag, 'r'))
data = {}
for row in csvreader1:
    if '#' not in row[0]:
        data[row[0]] = {'dcdft volume': float(row[1]),
                        'dcdft B0': float(row[2]),
                        'dcdft B1': float(row[3])}

csvwriter2 = csv.writer(open('%s.csv' % tag, 'wb'))
h2 = h + ['%' + h[1], '%' + h[2], '%' + h[3]]
csvwriter2.writerow(h2)

refs = np.loadtxt(reffile,
                  dtype={'names': ('element', 'V0', 'B0', 'BP'),
                         'formats': ('S2', float, float, float)})
# convert into dict
refsd = {}
for e, v, b0, b1 in refs:
    refsd[e] = [v, b0, b1]

rows = []
rowserr = []
for n in task.collection.names:
    row = [n]
    if n in data.keys():
        if 0:
            ref = task.collection.ref[n]  # don't use collection data
        else:
            ref = refsd[n]
        try:
            v = round(data[n]['dcdft volume'], 3)
            b0 = round(data[n]['dcdft B0'], 3)
            b1 = round(data[n]['dcdft B1'], 3)
            row.extend([v, b0, b1])
        except KeyError:  # completely failed to find eos minimum
            row.extend(['N/A', 'N/A', 'N/A'])
    else:
        # element not calculated
        row.extend(['N/A', 'N/A', 'N/A'])
    if 'N/A' not in row:
        v0, b00, b10 = ref
        ve = round((v - v0) / v0 * 100, 1)
        b0e = round((b0 - b00) / b00 * 100, 1)
        b1e = round((b1 - b10) / b10 * 100, 1)
        rows.append(row)
        # print row + ref + [ve, b0e, b1e]
        csvwriter2.writerow(row + [ve, b0e, b1e])

# calculate Delta
f = open('%s.txt' % tag, 'wb')
csvwriter3 = csv.writer(f, delimiter='\t')
for r in rows:
    csvwriter3.writerow(r)
f.close()
cmd = 'python ' + os.path.join(dir, 'calcDelta.py')
cmd += ' ' + '%s.txt ' % tag + reffile + ' --stdout'
cmd += ' > ' + '%s_Delta.txt' % tag
os.system(cmd)



r"""This module implements the linear operator in the Sternheimer equation.

Sternheimer equation::

       /           \             q
       | H - eps   | P      |dPsi  > = - P      dV  |Psi  > ,
       \        nk /  c,k+q      nk       c,k+q   q     nk

where P_c is the projection operator onto the unoccupied states. For a
perturbation with a specific q-vector, only the projections onto states at
k+q are needed.

"""

import numpy as np

from gpaw.utilities import unpack
from gpaw.utilities.blas import mmmx
from gpaw.fd_operators import Laplace


class SternheimerOperator:
    """Class implementing the linear operator in the Sternheimer equation.

    The main purpose of this class is to implement the multiplication of the
    linear operator (H - eps_nk) with a vector in the ``apply`` member
    function. An additional ``matvec`` member function has been defined so that
    instances of this class can be passed as a linear operator to scipy's
    iterative Krylov solvers in ``scipy.sparse.linalg``.

    """

    def __init__(self, hamiltonian, wfs, gd, dtype=float):
        """Save useful objects for the Sternheimer operator.

        Parameters
        ----------
        hamiltonian: Hamiltonian
            Hamiltonian for a ground-state calculation.
        wfs: WaveFunctions
            Ground-state wave-functions.
        gd: GridDescriptor
            Grid on which the operator is defined.
        dtype: dtype
            dtype of the wave-function being operated on.

        """

        self.hamiltonian = hamiltonian
        self.kin = Laplace(gd, scale=-0.5, n=3, dtype=dtype)
        self.kpt_u = wfs.kpt_u
        self.pt = wfs.pt
        self.gd = gd

        # Variables for k-point and band index
        self.k = None
        self.n = None
        self.kplusq = None

        # For scipy's linear solver
        N = np.prod(gd.n_c)
        self.shape = (N, N)
        self.dtype = dtype

    def set_k(self, k):
        """Set k-vector for the Bloch-state in consideration."""

        self.k = k

    def set_band(self, n):
        """Set band index for the Bloch-state in consideration.

        Note that n different from None has implications for the interpretation
        of the input vector in the ``apply`` member function.

        """

        self.n = n

    def set_kplusq(self, kplusq):
        """Set q-vector of the perturbing potential.

        Parameters
        ----------
        kplusq: int
           Index of the k+q vector.

        """

        self.kplusq = kplusq

    def apply(self, x_nG, y_nG):
        """Apply the linear operator in the Sternheimer equation to a vector.

        For the eigenvalue term the k-point is the one of the state.
        For the other terms the k-point to be used is the one given by the k+q
        phase of the first-order of the state. Only for q=0 do the two
        coincide.

        Parameters
        ----------
        x_nG: ndarray
            Vector(s) to which the Sternheimer operator is applied.
        y_nG: ndarray
            Resulting vector(s).

        """

        assert x_nG.ndim in (3, 4)
        assert x_nG.shape == y_nG.shape
        assert tuple(self.gd.n_c) == x_nG.shape[-3:]
        assert self.k is not None

        # k
        kpt = self.kpt_u[self.k]
        # k+q
        kplusqpt = self.kpt_u[self.kplusq]

        # Kintetic energy
        # k+q
        self.kin.apply(x_nG, y_nG, phase_cd=kplusqpt.phase_cd)

        # Local part of effective potential - no phase !!
        self.hamiltonian.apply_local_potential(x_nG, y_nG, kpt.s)

        # Non-local part from projectors (coefficients can not be reused)
        shape = x_nG.shape[:-3]
        P_ani = self.pt.dict(shape=shape)

        # k+q
        self.pt.integrate(x_nG, P_ani, q=kplusqpt.k)

        for a, P_ni in P_ani.items():
            dH_ii = unpack(self.hamiltonian.dH_asp[a][kpt.s])
            P_ani[a] = np.dot(P_ni, dH_ii)
        # k+q
        self.pt.add(y_nG, P_ani, q=kplusqpt.k)

        # XXX Eigenvalue term
        if self.n is not None:
            # k
            y_nG -= kpt.eps_n[self.n] * x_nG
        else:
            for n, x_G in enumerate(x_nG):
                # k
                y_nG[n] -= kpt.eps_n[n] * x_G

        # Project out undesired (numerical) components
        # k+q
        self.project(y_nG)

    def project(self, x_nG):
        """Project the vector onto the unoccupied states at k+q.

        The projection operator is defined as follows::

                      --                    --
             P      = >  |Psi ><Psi | = 1 - >  |Psi ><Psi |
              c,k+q   --     c     c        --     v     v
                       c    k+q   k+q        v    k+q   k+q

        """

        # It might be a good idea to move this functionality to its own class
        assert x_nG.ndim == 4
        assert self.kplusq is not None

        # k+q
        kplusqpt = self.kpt_u[self.kplusq]
        # Occupied wave function
        psit_nG = kplusqpt.psit_nG

        # Project out occupied sub-space using blas routine gemm
        m = len(x_nG)
        n = len(psit_nG)
        proj_mn = np.zeros((m, n), dtype=self.dtype)
        mmmx(self.gd.dv, x_nG, 'N', psit_nG, 'C', 0.0, proj_mn)
        mmmx(-1.0, proj_mn, 'N', psit_nG, 'N', 1.0, x_nG)

        # Project out one orbital at a time
        # for n, psit_G in enumerate(psit_nG):
        #
        #     proj = self.gd.integrate(psit_G.conjugate() * x_nG)
        #     x_nG -= proj * psit_G

    def matvec(self, x):
        """Matrix-vector multiplication for scipy's Krylov solvers.

        This is a wrapper around the ``apply`` member function above. It allows
        to multiply the sternheimer operator onto the 1-dimensional scipy
        representation of a gpaw grid vector(s).

        Parameters
        ----------
        a: ndarray
            1-dimensional array holding the representation of a gpaw grid
            vector (possibly a set of vectors).

        """

        # Check that a is 1-dimensional
        assert len(x.shape) == 1

        # Find the number of states in x
        grid_shape = tuple(self.gd.n_c)
        assert ((x.size % np.prod(grid_shape)) == 0), ("Incompatible array " +
                                                       "shapes")
        # Number of states in the vector
        N = x.size / np.prod(grid_shape)

        # Output array
        y_nG = self.gd.zeros(n=N, dtype=self.dtype)
        shape = y_nG.shape

        assert x.size == np.prod(y_nG.shape)

        x_nG = x.reshape(shape)

        self.apply(x_nG, y_nG)

        y = y_nG.ravel()

        return y



from math import factorial as fac
import numpy as np

from gpaw import debug
from gpaw.spline import Spline


def generate_bessel_coefficients(lmax):
    """Generate spherical Bessel function expansion coefficients.

    The coefficients c_lm (see expansion below) can be generated from the
    recurrence relation [see also master thesis of Marco Vanin (2008)]

                        2l+1
    j   (x) + j   (x) = ‾‾‾‾ j (x)    for l≥1
     l-1       l+1       x    l

    which implies that

    c     = (2l+1) c      for l≥1 and m∊{0,1}
     l+1m           lm

    while

    c     = (2l+1) c   - c          for l≥1 and m≥2
     l+1m           lm    l-1m-2

    With c_00 = i, c_10 = i and c_11=-1, it can be proven that the expansion
    has the closed-form solution:

           1+m     (2l-m)!
    c   = i    ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
     lm        2^(l-m)(l-m)!m!
    """
    c_lm = []
    for l in range(lmax + 1):
        c_m = []
        for m in range(l + 1):
            c_m.append((1.0j)**(1 + m) * fac(2 * l - m)
                       / (2**(l - m) * fac(l - m) * fac(m)))
        c_lm.append(c_m)
    return c_lm


c_lm = generate_bessel_coefficients(lmax=6)


def spherical_bessel(l, x_g):
    r"""Calculate the spherical Bessel function j_l(x).

    Evaluates the spherical Bessel function via the expansion [see master
    thesis of Marco Vanin (2008)]

                      l
                      __
                1     \
    j_l(x) = ‾‾‾‾‾‾‾  /  Re{ c_lm x^m e^(-ix) }
             x^(l+1)  ‾‾
                      m=0

    for non-negative real-valued x.
    """
    assert x_g.dtype == float and np.all(x_g >= 0)
    jl_g = np.zeros_like(x_g)

    # Mask out x = 0
    x0_g = x_g < 1e-10
    # j_0(x=0) = 1, j_l(x=0)=0 for l>0
    if l == 0:
        jl_g[x0_g] = 1.

    # Evaluate j_l(x) using the coefficients
    xpos_g = x_g[~x0_g]
    for m in range(l + 1):
        jl_g[~x0_g] += (c_lm[l][m] * xpos_g**m * np.exp(-1.j * xpos_g)).real
    jl_g[~x0_g] /= xpos_g**(l + 1)

    return jl_g


def ffbt(l, f_g, r_g, k_q):
    r"""Fast Fourier-Bessel transform.

    The following integral is calculated using l+1 FFTs::

                    oo
                   /
              l+1 |  2           l
      g(k) = k    | r dr j (kr) r f (r)
                  |       l
                 /
                  0

    Assuming that f(r)=0 ∀ r≥rc for some rc>0, the integral can be rewritten as
    follows using the coefficient expansion of j_l(x) given above:

           l
           __      /  rc                       \
           \   m   |  /      m+1       -ikr    |
    g(k) = /  k  Re<  | c   r    f(r) e     dr >
           ‾‾      |  /  lm                    |
           m=0     \  0                        /

    For a given uniform radial real-space grid,

    r(g) = g rc / N    for g=0,1,...,N-1

    we can evaluate the inner integral for values of k on a uniform radial
    reciprocal-space grid of length Q≥N,

            π
    k(q) = ‾‾‾‾ q    for q=0,1,...,Q-1
           Q Δr

    by applying the Fast-Fourier-Transform on a 2Q length grid, including also
    the corresponding "negative frequency" k-points:

    2Q-1
    __
    \   ⎧     m+1      ⎫|        -2πiqg/2Q
    /   |c   r    f(r) ||       e          Δr
    ‾‾  ⎩ lm           ⎭|
    g=0                  r=r(g)

    The "negative frequency" k-points are there only for the sake of the FFT
    and only the "positive frequency" k-points, corresponding to the radial
    reciprocal space grid, are returned.
    """

    dr = r_g[1]
    Q = len(k_q)

    if debug:
        # We assume a uniform real-space grid from r=0 to r=rc-Δr
        assert r_g[0] == 0.
        assert np.allclose(r_g[1:] - r_g[:-1], dr)
        # We assume a uniform reciprocal-space grid, with a specific grid
        # spacing in relation to Δr, starting from k=0 and including all the
        # "positive frequencies"
        assert k_q[0] == 0.
        assert Q >= len(r_g)
        dk = np.pi / (Q * dr)
        assert np.allclose(k_q[1:] - k_q[:-1], dk)

    # Perform the transform
    g_q = np.zeros(Q)
    for m in range(l + 1):
        g_q += (k_q**m *
                # Perform FFT with 2Q grid points (numpy automatically pads
                # the integrand with zeros for g≥N, that is for r≥rc)
                np.fft.fft(c_lm[l][m] * r_g**(m + 1) * f_g, 2 * Q)
                # Use the real part of the Q positive frequency points
                [:Q].real) * dr
    return g_q


def rescaled_bessel_limit(l):
    """Get the x->0 limit of a rescaled spherical bessel function.

    Calculates the closed form solution to the limit

         j_l(x)   2^l l!
    lim  ‾‾‾‾‾‾ = ‾‾‾‾‾‾‾
    x->0  x^l     (2l+1)!
    """
    return 2**l * fac(l) / fac(2 * l + 1)


class FourierBesselTransformer:
    def __init__(self, rcut, N):
        """Construct the transformer with ffbt-compliant grids."""
        # Set up radial real-space grid with N points
        self.N = N
        self.rcut = rcut
        self.dr = rcut / self.N
        self.r_g = np.arange(self.N) * self.dr
        # Set up radial reciprocal-space grid with Q=2N points
        self.Q = 2 * self.N
        self.dk = np.pi / (self.Q * self.dr)  # Use only "positive frequencies"
        self.k_q = np.arange(self.Q) * self.dk

    def transform(self, spline):
        """Fourier-Bessel transform a given radial function f(r).

        Calculates
                 rc
                 /
        f(k) = k | r^2 dr (kr)^l j_l(kr) f(r)
                 /
                 0
        """
        assert spline.get_cutoff() <= self.rcut, \
            f'Incompatible cutoffs {spline.get_cutoff()} and {self.rcut}'
        l = spline.get_angular_momentum_number()
        f_g = spline.map(self.r_g)
        f_q = ffbt(l, f_g, self.r_g, self.k_q)
        return f_q

    def rescaled_transform(self, spline):
        """Perform a rescaled Fourier-Bessel transform of f(r).

        Calculates

                  rc
        ˷         /         r^l
        f(k) = 4π | r^2 dr  ‾‾‾ j_l(kr) f(r)
                  /         k^l
                  0
        """
        # Calculate f(k) and rescale for finite k
        l = spline.get_angular_momentum_number()
        f_q = self.transform(spline)
        f_q[1:] *= 4 * np.pi / self.k_q[1:]**(2 * l + 1)

        # Calculate k=0 contribution
        f_q[0] = self.calculate_rescaled_average(spline)

        return f_q

    def calculate_rescaled_average(self, spline):
        """Calculate the rescaled transform for k=0.

        Calculates

                    rc
        ˷           /                     j_l(kr)
        f(k=0) = 4π | r^2 dr r^(2l)  lim  ‾‾‾‾‾‾‾ f(r)
                    /               kr->0 (kr)^l
                    0
        """
        l = spline.get_angular_momentum_number()
        f_g = spline.map(self.r_g)
        prefactor = 4 * np.pi * rescaled_bessel_limit(l) * self.dr
        return prefactor * np.dot(self.r_g**(2 * l + 2), f_g)


def rescaled_fourier_bessel_transform(spline, N=2**10):
    """Calculate rescaled Fourier-Bessel transform in spline representation."""
    # Fourier transform the spline, sampling it on a uniform grid
    rcut = 50.0  # Why not spline.get_cutoff() * 2 or similar?
    assert spline.get_cutoff() <= rcut
    transformer = FourierBesselTransformer(rcut, N)
    f_q = transformer.rescaled_transform(spline)

    # Return spline representation of the transform
    l = spline.get_angular_momentum_number()
    kmax = transformer.k_q[-1]
    return Spline(l, kmax, f_q)



from __future__ import annotations

import numpy as np
from gpaw.core.matrix import Matrix
from gpaw.external import ExternalPotential
from gpaw.lfc import BasisFunctions
from gpaw.new import zips
from gpaw.new.calculation import DFTState
from gpaw.new.fd.pot_calc import FDPotentialCalculator
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.new.lcao.wave_functions import LCAOWaveFunctions
from gpaw.typing import Array2D, Array3D


class HamiltonianMatrixCalculator:
    def calculate_matrix(self,
                         wfs: LCAOWaveFunctions) -> Matrix:
        raise NotImplementedError


class CollinearHamiltonianMatrixCalculator(HamiltonianMatrixCalculator):
    def __init__(self,
                 V_sxMM: list[np.ndarray],
                 dH_saii: list[dict[int, Array2D]],
                 basis: BasisFunctions,
                 include_kinetic: bool = True):
        self.V_sxMM = V_sxMM
        self.dH_saii = dH_saii
        self.basis = basis
        self.include_kinetic = include_kinetic

    def calculate_matrix(self,
                         wfs: LCAOWaveFunctions) -> Matrix:
        if self.include_kinetic:
            return self._calculate_matrix_with_kinetic(wfs)
        return self._calculate_matrix_without_kinetic(wfs)

    def _calculate_potential_matrix(self,
                                    wfs: LCAOWaveFunctions,
                                    V_xMM: Array3D) -> Matrix:
        data = V_xMM[0]
        _, M = data.shape
        if wfs.dtype == complex:
            data = data.astype(complex)
        V_MM = Matrix(M, M, data=data, dist=(wfs.band_comm, -1, 1))
        if wfs.dtype == complex:
            phase_x = np.exp(-2j * np.pi *
                             self.basis.sdisp_xc[1:] @ wfs.kpt_c)
            V_MM.data += np.einsum('x, xMN -> MN',
                                   2 * phase_x, V_xMM[1:],
                                   optimize=True)
        return V_MM

    def _calculate_matrix_without_kinetic(self,
                                          wfs: LCAOWaveFunctions,
                                          V_xMM: Array3D = None,
                                          dH_aii: dict[int, Array2D] = None
                                          ) -> Matrix:
        if V_xMM is None:
            V_xMM = self.V_sxMM[wfs.spin]
        if dH_aii is None:
            dH_aii = self.dH_saii[wfs.spin]

        V_MM = self._calculate_potential_matrix(wfs, V_xMM)

        M1, M2 = V_MM.dist.my_row_range()
        for a, dH_ii in dH_aii.items():
            P_Mi = wfs.P_aMi[a]
            V_MM.data += (P_Mi[M1:M2] @ dH_ii).conj() @ P_Mi.T  # XXX use gemm

        return V_MM

    def _calculate_matrix_with_kinetic(self,
                                       wfs: LCAOWaveFunctions) -> Matrix:
        H_MM = self._calculate_matrix_without_kinetic(wfs)
        H_MM.data += wfs.T_MM.data

        wfs.domain_comm.sum(H_MM.data, 0)

        if wfs.domain_comm.rank == 0:
            if wfs.dtype == complex:
                H_MM.add_hermitian_conjugate(scale=0.5)
            else:
                H_MM.tril2full()
        return H_MM


class NonCollinearHamiltonianMatrixCalculator(HamiltonianMatrixCalculator):
    def __init__(self, matcalc: CollinearHamiltonianMatrixCalculator):
        self.matcalc = matcalc

    def calculate_matrix(self,
                         wfs: LCAOWaveFunctions) -> Matrix:
        V_sMM = [
            self.matcalc._calculate_matrix_without_kinetic(wfs, V_xMM, dH_aii)
            for V_xMM, dH_aii in zips(self.matcalc.V_sxMM,
                                      self.matcalc.dH_saii)]

        V_sMM[0] += wfs.T_MM

        assert wfs.domain_comm.size == 1

        for V_MM in V_sMM:
            wfs.domain_comm.sum(V_MM.data, 0)

            if wfs.domain_comm.rank == 0:
                if wfs.dtype == complex:
                    V_MM.add_hermitian_conjugate(scale=0.5)
                else:
                    V_MM.tril2full()

        _, M = V_MM.shape
        v_MM, x_MM, y_MM, z_MM = (V_MM.data for V_MM in V_sMM)
        H_sMsM = Matrix(2 * M, 2 * M, dtype=complex, dist=(wfs.band_comm,))
        H_sMsM.data[:M, :M] = v_MM + z_MM
        H_sMsM.data[:M, M:] = x_MM + 1j * y_MM
        H_sMsM.data[M:, :M] = x_MM - 1j * y_MM
        H_sMsM.data[M:, M:] = v_MM - z_MM
        return H_sMsM


class LCAOHamiltonian(Hamiltonian):
    def __init__(self,
                 basis: BasisFunctions):
        self.basis = basis

    def create_hamiltonian_matrix_calculator(self,
                                             state: DFTState
                                             ) -> HamiltonianMatrixCalculator:
        V_sxMM = [self.basis.calculate_potential_matrices(vt_R.data)
                  for vt_R in state.potential.vt_sR.to_xp(np)]

        dH_saii = [{a: dH_sii[s]
                    for a, dH_sii
                    in state.potential.dH_asii.to_xp(np).items()}
                   for s in range(len(V_sxMM))]

        matcalc = CollinearHamiltonianMatrixCalculator(V_sxMM, dH_saii,
                                                       self.basis,
                                                       include_kinetic=True)
        if len(V_sxMM) < 4:
            return matcalc

        return NonCollinearHamiltonianMatrixCalculator(matcalc)

    def create_kick_matrix_calculator(self,
                                      state: DFTState,
                                      ext: ExternalPotential,
                                      pot_calc: FDPotentialCalculator
                                      ) -> HamiltonianMatrixCalculator:
        from gpaw.utilities import unpack
        vext_r = pot_calc.vbar_r.new()
        finegd = vext_r.desc._gd

        vext_r.data = ext.get_potential(finegd)
        vext_R = pot_calc.restrict(vext_r)

        nspins = state.ibzwfs.nspins

        V_MM = self.basis.calculate_potential_matrices(vext_R.data)
        V_sxMM = [V_MM for s in range(nspins)]

        W_aL = pot_calc.ghat_aLr.integrate(vext_r)

        assert state.ibzwfs.ibz.bz.gamma_only
        setups_a = state.ibzwfs.wfs_qs[0][0].setups

        dH_saii = [{a: unpack(setups_a[a].Delta_pL @ W_L)
                    for (a, W_L) in W_aL.items()}
                   for s in range(nspins)]

        return CollinearHamiltonianMatrixCalculator(V_sxMM, dH_saii,
                                                    self.basis,
                                                    include_kinetic=False)



import pytest
from ase.build import bulk
from gpaw import GPAW


def test_generic_si_primitive(in_tmp_dir):
    a = 5.475
    calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                kpts=(4, 4, 4),
                occupations={'name': 'tetrahedron-method'},
                spinpol=True,
                nbands=5)
    atoms = bulk('Si', 'diamond', a=a)
    atoms.calc = calc
    E = atoms.get_potential_energy()
    assert E == pytest.approx(-11.8584, abs=0.001)

    homo, lumo = calc.get_homo_lumo()
    assert lumo - homo == pytest.approx(1.117, abs=0.002)

    calc.write('si_primitive.gpw', 'all')
    calc = GPAW('si_primitive.gpw',
                parallel={'domain': 1, 'band': 1},
                txt=None)
